from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .reagent_ref import ReagentRef
from .simple_types import Color, NonNegativeInt, WellID
from .well_sample import WellSample


class Well(OMEType):
    """A Well is a component of the Well/Plate/Screen construct to describe screening
    applications.

    A Well has a number of WellSample elements that link to the Images collected
    in this well. The ReagentRef links any Reagents that were used in this Well. A
    well is part of only one Plate. The origin for the row and column identifiers
    is the top left corner of the plate starting at zero. i.e The top left well of
    a plate is index (0,0)

    Parameters
    ----------
    column : NonNegativeInt
        This is the column index of the well, the origin is the top left
        corner of the plate with the first column of cells being column zero.
        i.e top left is (0,0) The combination of Row, Column has to be unique
        for each well in a plate.
    id : WellID
    row : NonNegativeInt
        This is the row index of the well, the origin is the top left corner
        of the plate with the first row of wells being row zero. i.e top left
        is (0,0) The combination of Row, Column has to be unique for each well
        in a plate.
    annotation_ref : AnnotationRef, optional
    color : Color, optional
        A marker color used to highlight the well - encoded as RGBA The
        default value "-1" is #FFFFFFFF so solid white (it is a signed 32 bit
        value) NOTE: Prior to the 2012-06 schema the default value was
        incorrect and produced a transparent red not solid white.
    external_description : str, optional
        A description of the externally defined identifier for this plate.
    external_identifier : str, optional
        The ExternalIdentifier attribute may contain a reference to an
        external database.
    reagent_ref : ReagentRef, optional
    type : str, optional
        A human readable identifier for the screening status. e.g. empty,
        positive control, negative control, control, experimental, etc.
    well_samples : WellSample, optional
    """

    column: NonNegativeInt
    id: WellID
    row: NonNegativeInt
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    color: Optional[Color] = Color("white")
    external_description: Optional[str] = None
    external_identifier: Optional[str] = None
    reagent_ref: Optional[ReagentRef] = None
    type: Optional[str] = None
    well_samples: List[WellSample] = Field(default_factory=list)



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import AnnotationID


class AnnotationRef(Reference, OMEType):
    """The AnnotationRef element is a reference to an element derived from the
    CommonAnnotation element.

    Parameters
    ----------
    id : AnnotationID
    """

    id: AnnotationID



from enum import Enum
from typing import Optional

from typing_extensions import Literal

from ome_types._base_type import OMEType

from .light_source import LightSource


class Type(Enum):
    """The type of Arc lamp."""

    HG = "Hg"
    HG_XE = "HgXe"
    OTHER = "Other"
    XE = "Xe"


class Arc(LightSource, OMEType):
    """The Arc element is used to describe various kinds of Arc lamps - Hg, Xe, HgXe.

    The Power of the Arc is now stored in the LightSource.

    Parameters
    ----------
    id : LightSourceID
        A LightSource ID must be specified for each light source, and the
        individual light sources can be referred to by their LightSource IDs
        (eg from Channel).
    annotation_ref : AnnotationRef, optional
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    power : float, optional
        The light-source power. Units are set by PowerUnit.
    power_unit : UnitsPower, optional
        The units of the Power - default:milliwatts.
    serial_number : str, optional
        The serial number of the component.
    type : Type, optional
        The type of Arc lamp.
    """

    kind: Literal["arc"] = "arc"
    type: Optional[Type] = None



from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CustomBuildHook(BuildHookInterface):
    """Custom Plugin."""

    PLUGIN_NAME = "ome_autogen"

    def initialize(self, version: str, build_data: dict) -> None:
        """Init before the build process begins."""
        import sys

        sys.path.append("src")

        import ome_autogen

        ome_autogen.convert_schema()



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .dichroic_ref import DichroicRef
from .filter_ref import FilterRef
from .manufacturer_spec import ManufacturerSpec
from .simple_types import FilterSetID


class FilterSet(ManufacturerSpec, OMEType):
    """Filter set manufacturer specification

    Parameters
    ----------
    id : FilterSetID
    dichroic_ref : DichroicRef, optional
    emission_filter_ref : FilterRef, optional
        The Filters placed in the Emission light path.
    excitation_filter_ref : FilterRef, optional
        The Filters placed in the Excitation light path.
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    serial_number : str, optional
        The serial number of the component.
    """

    id: FilterSetID
    dichroic_ref: Optional[DichroicRef] = None
    emission_filter_ref: List[FilterRef] = Field(default_factory=list)
    excitation_filter_ref: List[FilterRef] = Field(default_factory=list)



from typing import Optional

from ome_types._base_type import OMEType


class Rights(OMEType):
    """The rights holder of this data and the rights held.

    Parameters
    ----------
    rights_held : str, optional
        The rights held by the rights holder. e.g. "All rights reserved" or
        "Creative Commons Attribution 3.0 Unported License"
    rights_holder : str, optional
        The rights holder for this data. e.g. "Copyright (C) 2002 - 2016 Open
        Microscopy Environment"
    """

    rights_held: Optional[str] = None
    rights_holder: Optional[str] = None



from enum import Enum
from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .manufacturer_spec import ManufacturerSpec
from .simple_types import FilterID
from .transmittance_range import TransmittanceRange


class Type(Enum):
    BAND_PASS = "BandPass"
    DICHROIC = "Dichroic"
    LONG_PASS = "LongPass"
    MULTI_PASS = "MultiPass"
    NEUTRAL_DENSITY = "NeutralDensity"
    OTHER = "Other"
    SHORT_PASS = "ShortPass"
    TUNEABLE = "Tuneable"


class Filter(ManufacturerSpec, OMEType):
    """A filter is either an excitation or emission filters.

    There should be one filter element specified per wavelength in the image. The
    channel number associated with a filter set is specified in Channel. It is
    based on the FilterSpec type, so has the required attributes Manufacturer,
    Model, and LotNumber. It may also contain a Type attribute which may be set to
    'LongPass', 'ShortPass', 'BandPass', 'MultiPass', 'Dichroic',
    'NeutralDensity', 'Tuneable' or 'Other'. It can be associated with an optional
    FilterWheel - Note: this is not the same as a FilterSet

    Parameters
    ----------
    id : FilterID
    annotation_ref : AnnotationRef, optional
    filter_wheel : str, optional
        A filter 'wheel' in OME can refer to any arrangement of filters in a
        filter holder of any shape. It could, for example, be a filter slider.
        [plain text string]
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    serial_number : str, optional
        The serial number of the component.
    transmittance_range : TransmittanceRange, optional
    type : Type, optional
    """

    id: FilterID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    filter_wheel: Optional[str] = None
    transmittance_range: Optional[TransmittanceRange] = None
    type: Optional[Type] = None



from __future__ import annotations

import re
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any

from . import model
from ._base_type import OMEType
from .model.reference import Reference

if TYPE_CHECKING:
    from .model.simple_types import LSID


def cast_number(qnum: str) -> str | int | float:
    """Attempt to cast a number from a string.

    This function attempts to cast a string to a number. It will first try to parse an
    int, then a float, and finally returns a string if both fail.
    """
    try:
        return int(qnum)
    except ValueError:
        try:
            return float(qnum)
        except ValueError:
            return qnum


def collect_references(value: Any) -> list[Reference]:
    """Return a list of all References contained in value.

    Recursively walks all dataclass fields and iterates over lists. The base
    case is when value is either a Reference object, or an uninteresting type
    that we don't need to inspect further.

    """
    references: list[Reference] = []
    if isinstance(value, Reference):
        references.append(value)
    elif isinstance(value, list):
        for v in value:
            references.extend(collect_references(v))
    elif isinstance(value, OMEType):
        for f in value.__fields__:
            references.extend(collect_references(getattr(value, f)))
    # Do nothing for uninteresting types
    return references


def collect_ids(value: Any) -> dict[LSID, OMEType]:
    """Return a map of all model objects contained in value, keyed by id.

    Recursively walks all dataclass fields and iterates over lists. The base
    case is when value is neither a dataclass nor a list.
    """
    ids: dict[LSID, OMEType] = {}
    if isinstance(value, list):
        for v in value:
            ids.update(collect_ids(v))
    elif isinstance(value, OMEType):
        for f in value.__fields__:
            if f == "id" and not isinstance(value, Reference):
                # We don't need to recurse on the id string, so just record it
                # and move on.
                ids[value.id] = value  # type: ignore
            else:
                ids.update(collect_ids(getattr(value, f)))
    # Do nothing for uninteresting types.
    return ids


CAMEL_REGEX = re.compile(r"(?<!^)(?=[A-Z])")


@lru_cache()
def camel_to_snake(name: str) -> str:
    """Return a snake_case version of a camelCase string."""
    return model._camel_to_snake.get(name, CAMEL_REGEX.sub("_", name).lower())


@lru_cache()
def norm_key(key: str) -> str:
    """Return a normalized key."""
    return key.split("}")[-1]


def _get_plural(key: str, tag: str) -> str:
    return model._singular_to_plural.get((norm_key(tag), key), key)


def _ensure_xml_bytes(path_or_str: Path | str | bytes) -> bytes:
    """Ensure that `path_or_str` is bytes.  Read from disk if it's an existing file."""
    if isinstance(path_or_str, Path):
        return path_or_str.read_bytes()
    if isinstance(path_or_str, str):
        # FIXME: deal with magic number 10.  I think it's to avoid Path.exists
        # failure on a full string
        if "xml" not in path_or_str[:10] and Path(path_or_str).exists():
            return Path(path_or_str).read_bytes()
        else:
            return path_or_str.encode()
    if isinstance(path_or_str, bytes):
        return path_or_str
    raise TypeError("path_or_str must be one of [Path, str, bytes].  I")



from enum import Enum
from typing import Optional

from ome_types._base_type import OMEType

from .simple_types import Hex40


class Compression(Enum):
    """Specifies the compression scheme used to encode the data."""

    BZIP2 = "bzip2"
    NONE = "none"
    ZLIB = "zlib"


class External(OMEType):
    """Describes a file location.

    Can optionally specify a portion of a file using Offset and a ReadLength.
    If Offset and ReadLength are specified in conjuction with Compression, then
    they point into the uncompressed file.

    Parameters
    ----------
    href : str
        file location
    sha1 : Hex40
        The digest of the file specified in href.
    compression : Compression, optional
        Specifies the compression scheme used to encode the data.
    """

    href: str
    sha1: Hex40
    compression: Optional[Compression] = Compression("none")



from typing import Union

from .ellipse import Ellipse
from .label import Label
from .line import Line
from .mask import Mask
from .point import Point
from .polygon import Polygon
from .polyline import Polyline
from .rectangle import Rectangle
from .shape import Shape

ShapeGroup = Shape

ShapeGroupType = Union[Rectangle, Mask, Point, Ellipse, Line, Polyline, Polygon, Label]



from napari_plugin_engine import napari_hook_implementation

from .widgets import OMETree

METADATA_KEY = "ome_types"


@napari_hook_implementation
def napari_experimental_provide_dock_widget():
    return OMETree, {"name": "OME Metadata Viewer"}


@napari_hook_implementation
def napari_get_reader(path):
    """Show OME XML if an ome.xml file is dropped on the viewer."""
    if isinstance(path, str) and path.endswith("ome.xml"):
        return view_ome_xml


def view_ome_xml(path):
    from napari._qt.qt_main_window import _QtMainWindow

    # close your eyes, or look away...
    # there is nothing worth looking at here!
    window = _QtMainWindow.current()
    if not window:
        return
    viewer = window.qt_viewer.viewer
    dw, widget = viewer.window.add_plugin_dock_widget("ome-types")
    widget.update(path)

    return [(None,)]  # sentinel



from enum import Enum
from typing import Optional

from ome_types._base_type import OMEType


class Compression(Enum):
    """Specifies the compression scheme used to encode the data."""

    BZIP2 = "bzip2"
    NONE = "none"
    ZLIB = "zlib"


class BinData(OMEType):
    """The contents of this element are base64-encoded.

    These are not CDATA sections, just a base64 stream.

    Parameters
    ----------
    big_endian : bool
        This is true if the binary data was written in BigEndian order. This
        is dependent on the system architecture of the machine that wrote the
        pixels. True for essentially all modern CPUs other than Intel and
        Alpha. All Binary data must be written in the same endian order.
    length : int
        Character count attribute for the BinData field. This is the length of
        the base-64 encoded block. It allows easy skipping of the block when
        parsing the file.
    compression : Compression, optional
        Specifies the compression scheme used to encode the data.
    """

    value: str
    big_endian: bool
    length: int
    compression: Optional[Compression] = Compression("none")



from ome_types._base_type import OMEType

from .annotation import Annotation
from .map import Map


class MapAnnotation(Annotation, OMEType):
    """An map annotation.

    The contents of this is a list of key/value pairs.

    Parameters
    ----------
    id : AnnotationID
    value : Map
    annotation_ref : AnnotationRef, optional
    annotator : ExperimenterID, optional
        The Annotator is the person who attached this annotation. e.g. If
        UserA annotates something with TagB, owned by UserB, UserA is still
        the Annotator.
    description : str, optional
        A description for the annotation.
    namespace : str, optional
        We recommend the inclusion of a namespace for annotations you define.
        If it is absent then we assume the annotation is to use our (OME's)
        default interpretation for this type.
    """

    value: Map



from .affine_transform import AffineTransform
from .annotation import Annotation
from .annotation_ref import AnnotationRef
from .arc import Arc
from .basic_annotation import BasicAnnotation
from .bin_data import BinData
from .binary_file import BinaryFile
from .boolean_annotation import BooleanAnnotation
from .channel import Channel
from .channel_ref import ChannelRef
from .comment_annotation import CommentAnnotation
from .dataset import Dataset
from .dataset_ref import DatasetRef
from .detector import Detector
from .detector_settings import DetectorSettings
from .dichroic import Dichroic
from .dichroic_ref import DichroicRef
from .double_annotation import DoubleAnnotation
from .ellipse import Ellipse
from .experiment import Experiment
from .experiment_ref import ExperimentRef
from .experimenter import Experimenter
from .experimenter_group import ExperimenterGroup
from .experimenter_group_ref import ExperimenterGroupRef
from .experimenter_ref import ExperimenterRef
from .external import External
from .filament import Filament
from .file_annotation import FileAnnotation
from .filter import Filter
from .filter_ref import FilterRef
from .filter_set import FilterSet
from .filter_set_ref import FilterSetRef
from .folder import Folder
from .folder_ref import FolderRef
from .generic_excitation_source import GenericExcitationSource
from .image import Image
from .image_ref import ImageRef
from .imaging_environment import ImagingEnvironment
from .instrument import Instrument
from .instrument_ref import InstrumentRef
from .label import Label
from .laser import Laser
from .leader import Leader
from .light_emitting_diode import LightEmittingDiode
from .light_path import LightPath
from .light_source import LightSource
from .light_source_group import LightSourceGroup
from .light_source_settings import LightSourceSettings
from .line import Line
from .list_annotation import ListAnnotation
from .long_annotation import LongAnnotation
from .manufacturer_spec import ManufacturerSpec
from .map import Map
from .map_annotation import MapAnnotation
from .mask import Mask
from .microbeam_manipulation import MicrobeamManipulation
from .microbeam_manipulation_ref import MicrobeamManipulationRef
from .microscope import Microscope
from .numeric_annotation import NumericAnnotation
from .objective import Objective
from .objective_settings import ObjectiveSettings
from .ome import OME
from .pixels import Pixels
from .plane import Plane
from .plate import Plate
from .plate_acquisition import PlateAcquisition
from .point import Point
from .polygon import Polygon
from .polyline import Polyline
from .project import Project
from .project_ref import ProjectRef
from .pump import Pump
from .reagent import Reagent
from .reagent_ref import ReagentRef
from .rectangle import Rectangle
from .reference import Reference
from .rights import Rights
from .roi import ROI
from .roi_ref import ROIRef
from .screen import Screen
from .settings import Settings
from .shape import Shape
from .shape_group import ShapeGroup
from .stage_label import StageLabel
from .structured_annotations import StructuredAnnotations
from .tag_annotation import TagAnnotation
from .term_annotation import TermAnnotation
from .text_annotation import TextAnnotation
from .tiff_data import TiffData
from .timestamp_annotation import TimestampAnnotation
from .transmittance_range import TransmittanceRange
from .type_annotation import TypeAnnotation
from .well import Well
from .well_sample import WellSample
from .well_sample_ref import WellSampleRef
from .xml_annotation import XMLAnnotation


__all__ = [
    "AffineTransform",
    "Annotation",
    "AnnotationRef",
    "Arc",
    "BasicAnnotation",
    "BinData",
    "BinaryFile",
    "BooleanAnnotation",
    "Channel",
    "ChannelRef",
    "CommentAnnotation",
    "Dataset",
    "DatasetRef",
    "Detector",
    "DetectorSettings",
    "Dichroic",
    "DichroicRef",
    "DoubleAnnotation",
    "Ellipse",
    "Experiment",
    "ExperimentRef",
    "Experimenter",
    "ExperimenterGroup",
    "ExperimenterGroupRef",
    "ExperimenterRef",
    "External",
    "Filament",
    "FileAnnotation",
    "Filter",
    "FilterRef",
    "FilterSet",
    "FilterSetRef",
    "Folder",
    "FolderRef",
    "GenericExcitationSource",
    "Image",
    "ImageRef",
    "ImagingEnvironment",
    "Instrument",
    "InstrumentRef",
    "Label",
    "Laser",
    "Leader",
    "LightEmittingDiode",
    "LightPath",
    "LightSource",
    "LightSourceGroup",
    "LightSourceSettings",
    "Line",
    "ListAnnotation",
    "LongAnnotation",
    "ManufacturerSpec",
    "Map",
    "MapAnnotation",
    "Mask",
    "MicrobeamManipulation",
    "MicrobeamManipulationRef",
    "Microscope",
    "NumericAnnotation",
    "OME",
    "Objective",
    "ObjectiveSettings",
    "Pixels",
    "Plane",
    "Plate",
    "PlateAcquisition",
    "Point",
    "Polygon",
    "Polyline",
    "Project",
    "ProjectRef",
    "Pump",
    "ROI",
    "ROIRef",
    "Reagent",
    "ReagentRef",
    "Rectangle",
    "Reference",
    "Rights",
    "Screen",
    "Settings",
    "Shape",
    "ShapeGroup",
    "StageLabel",
    "StructuredAnnotations",
    "TagAnnotation",
    "TermAnnotation",
    "TextAnnotation",
    "TiffData",
    "TimestampAnnotation",
    "TransmittanceRange",
    "TypeAnnotation",
    "Well",
    "WellSample",
    "WellSampleRef",
    "XMLAnnotation",
]

_singular_to_plural = {
    ("Experiment", "microbeam_manipulation"): "microbeam_manipulations",
    ("Instrument", "detector"): "detectors",
    ("Instrument", "dichroic"): "dichroics",
    ("Instrument", "filter"): "filters",
    ("Instrument", "filter_set"): "filter_sets",
    ("Instrument", "objective"): "objectives",
    ("OME", "dataset"): "datasets",
    ("OME", "experiment"): "experiments",
    ("OME", "experimenter"): "experimenters",
    ("OME", "experimenter_group"): "experimenter_groups",
    ("OME", "folder"): "folders",
    ("OME", "image"): "images",
    ("OME", "instrument"): "instruments",
    ("OME", "plate"): "plates",
    ("OME", "project"): "projects",
    ("OME", "roi"): "rois",
    ("OME", "screen"): "screens",
    ("Pixels", "channel"): "channels",
    ("Pixels", "plane"): "planes",
    ("Pixels", "tiff_data"): "tiff_data_blocks",
    ("Plate", "plate_acquisition"): "plate_acquisitions",
    ("Plate", "well"): "wells",
    ("Screen", "reagent"): "reagents",
    ("StructuredAnnotations", "boolean_annotation"): "boolean_annotations",
    ("StructuredAnnotations", "comment_annotation"): "comment_annotations",
    ("StructuredAnnotations", "double_annotation"): "double_annotations",
    ("StructuredAnnotations", "file_annotation"): "file_annotations",
    ("StructuredAnnotations", "list_annotation"): "list_annotations",
    ("StructuredAnnotations", "long_annotation"): "long_annotations",
    ("StructuredAnnotations", "map_annotation"): "map_annotations",
    ("StructuredAnnotations", "tag_annotation"): "tag_annotations",
    ("StructuredAnnotations", "term_annotation"): "term_annotations",
    ("StructuredAnnotations", "timestamp_annotation"): "timestamp_annotations",
    ("StructuredAnnotations", "xml_annotation"): "xml_annotations",
    ("Well", "well_sample"): "well_samples",
}

_plural_to_singular = {
    "microbeam_manipulations": "microbeam_manipulation",
    "detectors": "detector",
    "dichroics": "dichroic",
    "filters": "filter",
    "filter_sets": "filter_set",
    "objectives": "objective",
    "datasets": "dataset",
    "experiments": "experiment",
    "experimenters": "experimenter",
    "experimenter_groups": "experimenter_group",
    "folders": "folder",
    "images": "image",
    "instruments": "instrument",
    "plates": "plate",
    "projects": "project",
    "rois": "roi",
    "screens": "screen",
    "channels": "channel",
    "planes": "plane",
    "tiff_data_blocks": "tiff_data",
    "plate_acquisitions": "plate_acquisition",
    "wells": "well",
    "reagents": "reagent",
    "boolean_annotations": "boolean_annotation",
    "comment_annotations": "comment_annotation",
    "double_annotations": "double_annotation",
    "file_annotations": "file_annotation",
    "list_annotations": "list_annotation",
    "long_annotations": "long_annotation",
    "map_annotations": "map_annotation",
    "tag_annotations": "tag_annotation",
    "term_annotations": "term_annotation",
    "timestamp_annotations": "timestamp_annotation",
    "xml_annotations": "xml_annotation",
    "well_samples": "well_sample",
}

_camel_to_snake = {
    "A00": "a00",
    "A01": "a01",
    "A02": "a02",
    "A10": "a10",
    "A11": "a11",
    "A12": "a12",
    "APD": "apd",
    "Achro": "achro",
    "Achromat": "achromat",
    "AcquisitionDate": "acquisition_date",
    "AcquisitionMode": "acquisition_mode",
    "AffineTransform": "affine_transform",
    "Ag": "ag",
    "Air": "air",
    "AirPressure": "air_pressure",
    "AirPressureUnit": "air_pressure_unit",
    "Alexandrite": "alexandrite",
    "AmplificationGain": "amplification_gain",
    "AnalogVideo": "analog_video",
    "Annotation": "annotation",
    "AnnotationRef": "annotation_ref",
    "Annotator": "annotator",
    "Apo": "apo",
    "Ar": "ar",
    "ArCl": "ar_cl",
    "ArFl": "ar_fl",
    "Arc": "arc",
    "Arrow": "arrow",
    "Attenuation": "attenuation",
    "BandPass": "band_pass",
    "BasicAnnotation": "basic_annotation",
    "BigEndian": "big_endian",
    "BinData": "bin_data",
    "BinaryFile": "binary_file",
    "BinaryOnly": "binary_only",
    "Binning": "binning",
    "Bold": "bold",
    "BoldItalic": "bold_italic",
    "BooleanAnnotation": "boolean_annotation",
    "BooleanAnnotations": "boolean_annotations",
    "BrightField": "bright_field",
    "Brightfield": "brightfield",
    "CCD": "ccd",
    "CMOS": "cmos",
    "CO": "co",
    "CO2": "co2",
    "CO2Percent": "co2_percent",
    "CW": "cw",
    "CalibratedMagnification": "calibrated_magnification",
    "Channel": "channel",
    "ChannelRef": "channel_ref",
    "Channels": "channels",
    "Colocalization": "colocalization",
    "Color": "color",
    "Column": "column",
    "ColumnNamingConvention": "column_naming_convention",
    "Columns": "columns",
    "CommentAnnotation": "comment_annotation",
    "CommentAnnotations": "comment_annotations",
    "Compression": "compression",
    "ContrastMethod": "contrast_method",
    "Correction": "correction",
    "CorrectionCollar": "correction_collar",
    "CorrelationSpectroscopy": "correlation_spectroscopy",
    "CoumarinC30": "coumarin_c30",
    "Creator": "creator",
    "Cu": "cu",
    "CutIn": "cut_in",
    "CutInTolerance": "cut_in_tolerance",
    "CutInToleranceUnit": "cut_in_tolerance_unit",
    "CutInUnit": "cut_in_unit",
    "CutOut": "cut_out",
    "CutOutTolerance": "cut_out_tolerance",
    "CutOutToleranceUnit": "cut_out_tolerance_unit",
    "CutOutUnit": "cut_out_unit",
    "DIC": "dic",
    "Darkfield": "darkfield",
    "Dataset": "dataset",
    "DatasetRef": "dataset_ref",
    "Datasets": "datasets",
    "DeltaT": "delta_t",
    "DeltaTUnit": "delta_t_unit",
    "Description": "description",
    "Detector": "detector",
    "DetectorSettings": "detector_settings",
    "Detectors": "detectors",
    "Dichroic": "dichroic",
    "DichroicRef": "dichroic_ref",
    "Dichroics": "dichroics",
    "DimensionOrder": "dimension_order",
    "Dissection": "dissection",
    "DoubleAnnotation": "double_annotation",
    "DoubleAnnotations": "double_annotations",
    "Dye": "dye",
    "EBCCD": "ebccd",
    "EMCCD": "emccd",
    "EMinus": "e_minus",
    "Electrophysiology": "electrophysiology",
    "Ellipse": "ellipse",
    "Email": "email",
    "EmissionFilterRef": "emission_filter_ref",
    "EmissionWavelength": "emission_wavelength",
    "EmissionWavelengthUnit": "emission_wavelength_unit",
    "EndTime": "end_time",
    "Epifluorescence": "epifluorescence",
    "ErGlass": "er_glass",
    "ErYAG": "er_yag",
    "EvenOdd": "even_odd",
    "Excimer": "excimer",
    "ExcitationFilterRef": "excitation_filter_ref",
    "ExcitationWavelength": "excitation_wavelength",
    "ExcitationWavelengthUnit": "excitation_wavelength_unit",
    "Experiment": "experiment",
    "ExperimentRef": "experiment_ref",
    "Experimenter": "experimenter",
    "ExperimenterGroup": "experimenter_group",
    "ExperimenterGroupRef": "experimenter_group_ref",
    "ExperimenterGroups": "experimenter_groups",
    "ExperimenterRef": "experimenter_ref",
    "Experimenters": "experimenters",
    "Experiments": "experiments",
    "ExposureTime": "exposure_time",
    "ExposureTimeUnit": "exposure_time_unit",
    "External": "external",
    "ExternalDescription": "external_description",
    "ExternalIdentifier": "external_identifier",
    "FISH": "fish",
    "FLIP": "flip",
    "FP": "fp",
    "FRAP": "frap",
    "FRET": "fret",
    "FSM": "fsm",
    "FTIR": "ftir",
    "FieldIndex": "field_index",
    "Filament": "filament",
    "FileAnnotation": "file_annotation",
    "FileAnnotations": "file_annotations",
    "FileName": "file_name",
    "FillColor": "fill_color",
    "FillRule": "fill_rule",
    "Filter": "filter",
    "FilterRef": "filter_ref",
    "FilterSet": "filter_set",
    "FilterSetRef": "filter_set_ref",
    "FilterSets": "filter_sets",
    "FilterWheel": "filter_wheel",
    "Filters": "filters",
    "FirstC": "first_c",
    "FirstName": "first_name",
    "FirstT": "first_t",
    "FirstZ": "first_z",
    "Fl": "fl",
    "Fluar": "fluar",
    "Fluor": "fluor",
    "Fluorescence": "fluorescence",
    "FluorescenceCorrelationSpectroscopy": "fluorescence_correlation_spectroscopy",
    "FluorescenceLifetime": "fluorescence_lifetime",
    "Fluotar": "fluotar",
    "Folder": "folder",
    "FolderRef": "folder_ref",
    "Folders": "folders",
    "FontFamily": "font_family",
    "FontSize": "font_size",
    "FontSizeUnit": "font_size_unit",
    "FontStyle": "font_style",
    "FourDPlus": "four_d_plus",
    "FreeElectron": "free_electron",
    "FrequencyMultiplication": "frequency_multiplication",
    "GaAlAs": "ga_al_as",
    "GaAs": "ga_as",
    "Gain": "gain",
    "Gas": "gas",
    "GenericExcitationSource": "generic_excitation_source",
    "Glycerol": "glycerol",
    "H2O": "h2_o",
    "HFl": "h_fl",
    "Halogen": "halogen",
    "HashSHA1": "hash_sha1",
    "HeCd": "he_cd",
    "HeNe": "he_ne",
    "Height": "height",
    "Hg": "hg",
    "HgXe": "hg_xe",
    "HoYAG": "ho_yag",
    "HoYLF": "ho_ylf",
    "HoffmanModulation": "hoffman_modulation",
    "Humidity": "humidity",
    "ID": "id",
    "IFD": "ifd",
    "IlluminationType": "illumination_type",
    "Image": "image",
    "ImageRef": "image_ref",
    "Images": "images",
    "ImagingEnvironment": "imaging_environment",
    "Immersion": "immersion",
    "Immunocytochemistry": "immunocytochemistry",
    "Immunofluorescence": "immunofluorescence",
    "Incandescent": "incandescent",
    "Index": "index",
    "Institution": "institution",
    "Instrument": "instrument",
    "InstrumentRef": "instrument_ref",
    "Instruments": "instruments",
    "Integration": "integration",
    "IntensifiedCCD": "intensified_ccd",
    "Interleaved": "interleaved",
    "InverseFRAP": "inverse_frap",
    "Inverted": "inverted",
    "IonImaging": "ion_imaging",
    "Iris": "iris",
    "Italic": "italic",
    "K": "k",
    "Kr": "kr",
    "KrCl": "kr_cl",
    "KrFl": "kr_fl",
    "LCM": "lcm",
    "Label": "label",
    "Laser": "laser",
    "LaserMedium": "laser_medium",
    "LaserScanningConfocalMicroscopy": "laser_scanning_confocal_microscopy",
    "LastName": "last_name",
    "Leader": "leader",
    "Length": "length",
    "LensNA": "lens_na",
    "LifetimeImaging": "lifetime_imaging",
    "LightEmittingDiode": "light_emitting_diode",
    "LightPath": "light_path",
    "LightSource": "light_source",
    "LightSourceGroup": "light_source_group",
    "LightSourceSettings": "light_source_settings",
    "Line": "line",
    "ListAnnotation": "list_annotation",
    "ListAnnotations": "list_annotations",
    "Locked": "locked",
    "LongAnnotation": "long_annotation",
    "LongAnnotations": "long_annotations",
    "LongPass": "long_pass",
    "LotNumber": "lot_number",
    "M": "m",
    "MIMEType": "mime_type",
    "Manufacturer": "manufacturer",
    "ManufacturerSpec": "manufacturer_spec",
    "Map": "map",
    "MapAnnotation": "map_annotation",
    "MapAnnotations": "map_annotations",
    "MarkerEnd": "marker_end",
    "MarkerStart": "marker_start",
    "Mask": "mask",
    "MaximumFieldCount": "maximum_field_count",
    "Medium": "medium",
    "MetadataFile": "metadata_file",
    "MetadataOnly": "metadata_only",
    "MetalVapor": "metal_vapor",
    "MicrobeamManipulation": "microbeam_manipulation",
    "MicrobeamManipulationRef": "microbeam_manipulation_ref",
    "MicrobeamManipulations": "microbeam_manipulations",
    "Microscope": "microscope",
    "MiddleName": "middle_name",
    "ModeLocked": "mode_locked",
    "Model": "model",
    "Multi": "multi",
    "MultiPass": "multi_pass",
    "MultiPhotonMicroscopy": "multi_photon_microscopy",
    "N": "n",
    "NDFilter": "nd_filter",
    "Name": "name",
    "Namespace": "namespace",
    "NdGlass": "nd_glass",
    "NdYAG": "nd_yag",
    "NearFieldScanningOpticalMicroscopy": "near_field_scanning_optical_microscopy",
    "Neofluar": "neofluar",
    "NeutralDensity": "neutral_density",
    "NominalMagnification": "nominal_magnification",
    "NonLinear": "non_linear",
    "NonZero": "non_zero",
    "Normal": "normal",
    "NumericAnnotation": "numeric_annotation",
    "OME": "ome",
    "Objective": "objective",
    "ObjectiveSettings": "objective_settings",
    "Objectives": "objectives",
    "Oblique": "oblique",
    "ObliqueIllumination": "oblique_illumination",
    "Offset": "offset",
    "Oil": "oil",
    "OpticalTrapping": "optical_trapping",
    "Other": "other",
    "PALM": "palm",
    "PGIDocumentation": "pgi_documentation",
    "PMT": "pmt",
    "Phase": "phase",
    "Photoablation": "photoablation",
    "Photoactivation": "photoactivation",
    "Photobleaching": "photobleaching",
    "Photodiode": "photodiode",
    "PhysicalSizeX": "physical_size_x",
    "PhysicalSizeXUnit": "physical_size_x_unit",
    "PhysicalSizeY": "physical_size_y",
    "PhysicalSizeYUnit": "physical_size_y_unit",
    "PhysicalSizeZ": "physical_size_z",
    "PhysicalSizeZUnit": "physical_size_z_unit",
    "PinholeSize": "pinhole_size",
    "PinholeSizeUnit": "pinhole_size_unit",
    "Pixels": "pixels",
    "PlanApo": "plan_apo",
    "PlanFluor": "plan_fluor",
    "PlanNeofluar": "plan_neofluar",
    "Plane": "plane",
    "PlaneCount": "plane_count",
    "Planes": "planes",
    "Plate": "plate",
    "PlateAcquisition": "plate_acquisition",
    "PlateAcquisitions": "plate_acquisitions",
    "PlateRef": "plate_ref",
    "Plates": "plates",
    "PockelCell": "pockel_cell",
    "PockelCellSetting": "pockel_cell_setting",
    "Point": "point",
    "Points": "points",
    "PolarizedLight": "polarized_light",
    "Polygon": "polygon",
    "Polyline": "polyline",
    "PositionX": "position_x",
    "PositionXUnit": "position_x_unit",
    "PositionY": "position_y",
    "PositionYUnit": "position_y_unit",
    "PositionZ": "position_z",
    "PositionZUnit": "position_z_unit",
    "Power": "power",
    "PowerUnit": "power_unit",
    "Project": "project",
    "ProjectRef": "project_ref",
    "Projects": "projects",
    "ProtocolDescription": "protocol_description",
    "ProtocolIdentifier": "protocol_identifier",
    "Pulse": "pulse",
    "Pump": "pump",
    "QSwitched": "q_switched",
    "ROI": "roi",
    "ROIRef": "roi_ref",
    "ROIs": "rois",
    "RadiusX": "radius_x",
    "RadiusY": "radius_y",
    "ReadOutRate": "read_out_rate",
    "ReadOutRateUnit": "read_out_rate_unit",
    "Reagent": "reagent",
    "ReagentIdentifier": "reagent_identifier",
    "ReagentRef": "reagent_ref",
    "ReagentSetDescription": "reagent_set_description",
    "ReagentSetIdentifier": "reagent_set_identifier",
    "Reagents": "reagents",
    "Rectangle": "rectangle",
    "Reference": "reference",
    "RefractiveIndex": "refractive_index",
    "RepetitionRate": "repetition_rate",
    "RepetitionRateUnit": "repetition_rate_unit",
    "Repetitive": "repetitive",
    "Rhodamine6G": "rhodamine6_g",
    "Rights": "rights",
    "RightsHeld": "rights_held",
    "RightsHolder": "rights_holder",
    "Row": "row",
    "RowNamingConvention": "row_naming_convention",
    "Rows": "rows",
    "Ruby": "ruby",
    "SHA1": "sha1",
    "SPIM": "spim",
    "STED": "sted",
    "STORM": "storm",
    "SamplesPerPixel": "samples_per_pixel",
    "Screen": "screen",
    "Screens": "screens",
    "SecondHarmonicGenerationImaging": "second_harmonic_generation_imaging",
    "Semiconductor": "semiconductor",
    "SerialNumber": "serial_number",
    "Settings": "settings",
    "Shape": "shape",
    "ShapeGroup": "shape_group",
    "ShortPass": "short_pass",
    "SignificantBits": "significant_bits",
    "Single": "single",
    "SingleMoleculeImaging": "single_molecule_imaging",
    "Size": "size",
    "SizeC": "size_c",
    "SizeT": "size_t",
    "SizeX": "size_x",
    "SizeY": "size_y",
    "SizeZ": "size_z",
    "SlitScanConfocal": "slit_scan_confocal",
    "SolidState": "solid_state",
    "SpectralImaging": "spectral_imaging",
    "Spectroscopy": "spectroscopy",
    "SpinningDiskConfocal": "spinning_disk_confocal",
    "StageLabel": "stage_label",
    "StartTime": "start_time",
    "Status": "status",
    "StrokeColor": "stroke_color",
    "StrokeDashArray": "stroke_dash_array",
    "StrokeWidth": "stroke_width",
    "StrokeWidthUnit": "stroke_width_unit",
    "StructuredAnnotations": "structured_annotations",
    "StructuredIllumination": "structured_illumination",
    "SuperFluor": "super_fluor",
    "SweptFieldConfocal": "swept_field_confocal",
    "TIRF": "tirf",
    "TagAnnotation": "tag_annotation",
    "TagAnnotations": "tag_annotations",
    "Temperature": "temperature",
    "TemperatureUnit": "temperature_unit",
    "TermAnnotation": "term_annotation",
    "TermAnnotations": "term_annotations",
    "Text": "text",
    "TextAnnotation": "text_annotation",
    "TheC": "the_c",
    "TheT": "the_t",
    "TheZ": "the_z",
    "TiSapphire": "ti_sapphire",
    "TiffData": "tiff_data",
    "TiffDataBlocks": "tiff_data_blocks",
    "TimeIncrement": "time_increment",
    "TimeIncrementUnit": "time_increment_unit",
    "TimeLapse": "time_lapse",
    "Timepoint": "timepoint",
    "TimestampAnnotation": "timestamp_annotation",
    "TimestampAnnotations": "timestamp_annotations",
    "TotalInternalReflection": "total_internal_reflection",
    "Transform": "transform",
    "Transmittance": "transmittance",
    "TransmittanceRange": "transmittance_range",
    "Transmitted": "transmitted",
    "Tuneable": "tuneable",
    "Type": "type",
    "TypeAnnotation": "type_annotation",
    "UUID": "uuid",
    "UV": "uv",
    "Uncaging": "uncaging",
    "Union": "union",
    "Upright": "upright",
    "UserName": "user_name",
    "Value": "value",
    "VioletCorrected": "violet_corrected",
    "Voltage": "voltage",
    "VoltageUnit": "voltage_unit",
    "Water": "water",
    "WaterDipping": "water_dipping",
    "Wavelength": "wavelength",
    "WavelengthUnit": "wavelength_unit",
    "Well": "well",
    "WellOriginX": "well_origin_x",
    "WellOriginXUnit": "well_origin_x_unit",
    "WellOriginY": "well_origin_y",
    "WellOriginYUnit": "well_origin_y_unit",
    "WellSample": "well_sample",
    "WellSampleRef": "well_sample_ref",
    "WellSamples": "well_samples",
    "Wells": "wells",
    "WideField": "wide_field",
    "Width": "width",
    "WorkingDistance": "working_distance",
    "WorkingDistanceUnit": "working_distance_unit",
    "X": "x",
    "X1": "x1",
    "X2": "x2",
    "XMLAnnotation": "xml_annotation",
    "XMLAnnotations": "xml_annotations",
    "XUnit": "x_unit",
    "XYCTZ": "xyctz",
    "XYCZT": "xyczt",
    "XYTCZ": "xytcz",
    "XYTZC": "xytzc",
    "XYZCT": "xyzct",
    "XYZTC": "xyztc",
    "Xe": "xe",
    "XeBr": "xe_br",
    "XeCl": "xe_cl",
    "XeFl": "xe_fl",
    "Y": "y",
    "Y1": "y1",
    "Y2": "y2",
    "YUnit": "y_unit",
    "Z": "z",
    "ZUnit": "z_unit",
    "Zoom": "zoom",
    "bzip2": "bzip2",
    "cursive": "cursive",
    "fantasy": "fantasy",
    "href": "href",
    "letter": "letter",
    "monospace": "monospace",
    "none": "none",
    "number": "number",
    "sans-serif": "sans-serif",
    "serif": "serif",
    "zlib": "zlib",
}

_snake_to_camel = {
    "a00": "A00",
    "a01": "A01",
    "a02": "A02",
    "a10": "A10",
    "a11": "A11",
    "a12": "A12",
    "apd": "APD",
    "achro": "Achro",
    "achromat": "Achromat",
    "acquisition_date": "AcquisitionDate",
    "acquisition_mode": "AcquisitionMode",
    "affine_transform": "AffineTransform",
    "ag": "Ag",
    "air": "Air",
    "air_pressure": "AirPressure",
    "air_pressure_unit": "AirPressureUnit",
    "alexandrite": "Alexandrite",
    "amplification_gain": "AmplificationGain",
    "analog_video": "AnalogVideo",
    "annotation": "Annotation",
    "annotation_ref": "AnnotationRef",
    "annotator": "Annotator",
    "apo": "Apo",
    "ar": "Ar",
    "ar_cl": "ArCl",
    "ar_fl": "ArFl",
    "arc": "Arc",
    "arrow": "Arrow",
    "attenuation": "Attenuation",
    "band_pass": "BandPass",
    "basic_annotation": "BasicAnnotation",
    "big_endian": "BigEndian",
    "bin_data": "BinData",
    "binary_file": "BinaryFile",
    "binary_only": "BinaryOnly",
    "binning": "Binning",
    "bold": "Bold",
    "bold_italic": "BoldItalic",
    "boolean_annotation": "BooleanAnnotation",
    "boolean_annotations": "BooleanAnnotations",
    "bright_field": "BrightField",
    "brightfield": "Brightfield",
    "ccd": "CCD",
    "cmos": "CMOS",
    "co": "CO",
    "co2": "CO2",
    "co2_percent": "CO2Percent",
    "cw": "CW",
    "calibrated_magnification": "CalibratedMagnification",
    "channel": "Channel",
    "channel_ref": "ChannelRef",
    "channels": "Channels",
    "colocalization": "Colocalization",
    "color": "Color",
    "column": "Column",
    "column_naming_convention": "ColumnNamingConvention",
    "columns": "Columns",
    "comment_annotation": "CommentAnnotation",
    "comment_annotations": "CommentAnnotations",
    "compression": "Compression",
    "contrast_method": "ContrastMethod",
    "correction": "Correction",
    "correction_collar": "CorrectionCollar",
    "correlation_spectroscopy": "CorrelationSpectroscopy",
    "coumarin_c30": "CoumarinC30",
    "creator": "Creator",
    "cu": "Cu",
    "cut_in": "CutIn",
    "cut_in_tolerance": "CutInTolerance",
    "cut_in_tolerance_unit": "CutInToleranceUnit",
    "cut_in_unit": "CutInUnit",
    "cut_out": "CutOut",
    "cut_out_tolerance": "CutOutTolerance",
    "cut_out_tolerance_unit": "CutOutToleranceUnit",
    "cut_out_unit": "CutOutUnit",
    "dic": "DIC",
    "darkfield": "Darkfield",
    "dataset": "Dataset",
    "dataset_ref": "DatasetRef",
    "datasets": "Datasets",
    "delta_t": "DeltaT",
    "delta_t_unit": "DeltaTUnit",
    "description": "Description",
    "detector": "Detector",
    "detector_settings": "DetectorSettings",
    "detectors": "Detectors",
    "dichroic": "Dichroic",
    "dichroic_ref": "DichroicRef",
    "dichroics": "Dichroics",
    "dimension_order": "DimensionOrder",
    "dissection": "Dissection",
    "double_annotation": "DoubleAnnotation",
    "double_annotations": "DoubleAnnotations",
    "dye": "Dye",
    "ebccd": "EBCCD",
    "emccd": "EMCCD",
    "e_minus": "EMinus",
    "electrophysiology": "Electrophysiology",
    "ellipse": "Ellipse",
    "email": "Email",
    "emission_filter_ref": "EmissionFilterRef",
    "emission_wavelength": "EmissionWavelength",
    "emission_wavelength_unit": "EmissionWavelengthUnit",
    "end_time": "EndTime",
    "epifluorescence": "Epifluorescence",
    "er_glass": "ErGlass",
    "er_yag": "ErYAG",
    "even_odd": "EvenOdd",
    "excimer": "Excimer",
    "excitation_filter_ref": "ExcitationFilterRef",
    "excitation_wavelength": "ExcitationWavelength",
    "excitation_wavelength_unit": "ExcitationWavelengthUnit",
    "experiment": "Experiment",
    "experiment_ref": "ExperimentRef",
    "experimenter": "Experimenter",
    "experimenter_group": "ExperimenterGroup",
    "experimenter_group_ref": "ExperimenterGroupRef",
    "experimenter_groups": "ExperimenterGroups",
    "experimenter_ref": "ExperimenterRef",
    "experimenters": "Experimenters",
    "experiments": "Experiments",
    "exposure_time": "ExposureTime",
    "exposure_time_unit": "ExposureTimeUnit",
    "external": "External",
    "external_description": "ExternalDescription",
    "external_identifier": "ExternalIdentifier",
    "fish": "FISH",
    "flip": "FLIP",
    "fp": "FP",
    "frap": "FRAP",
    "fret": "FRET",
    "fsm": "FSM",
    "ftir": "FTIR",
    "field_index": "FieldIndex",
    "filament": "Filament",
    "file_annotation": "FileAnnotation",
    "file_annotations": "FileAnnotations",
    "file_name": "FileName",
    "fill_color": "FillColor",
    "fill_rule": "FillRule",
    "filter": "Filter",
    "filter_ref": "FilterRef",
    "filter_set": "FilterSet",
    "filter_set_ref": "FilterSetRef",
    "filter_sets": "FilterSets",
    "filter_wheel": "FilterWheel",
    "filters": "Filters",
    "first_c": "FirstC",
    "first_name": "FirstName",
    "first_t": "FirstT",
    "first_z": "FirstZ",
    "fl": "Fl",
    "fluar": "Fluar",
    "fluor": "Fluor",
    "fluorescence": "Fluorescence",
    "fluorescence_correlation_spectroscopy": "FluorescenceCorrelationSpectroscopy",
    "fluorescence_lifetime": "FluorescenceLifetime",
    "fluotar": "Fluotar",
    "folder": "Folder",
    "folder_ref": "FolderRef",
    "folders": "Folders",
    "font_family": "FontFamily",
    "font_size": "FontSize",
    "font_size_unit": "FontSizeUnit",
    "font_style": "FontStyle",
    "four_d_plus": "FourDPlus",
    "free_electron": "FreeElectron",
    "frequency_multiplication": "FrequencyMultiplication",
    "ga_al_as": "GaAlAs",
    "ga_as": "GaAs",
    "gain": "Gain",
    "gas": "Gas",
    "generic_excitation_source": "GenericExcitationSource",
    "glycerol": "Glycerol",
    "h2_o": "H2O",
    "h_fl": "HFl",
    "halogen": "Halogen",
    "hash_sha1": "HashSHA1",
    "he_cd": "HeCd",
    "he_ne": "HeNe",
    "height": "Height",
    "hg": "Hg",
    "hg_xe": "HgXe",
    "ho_yag": "HoYAG",
    "ho_ylf": "HoYLF",
    "hoffman_modulation": "HoffmanModulation",
    "humidity": "Humidity",
    "id": "ID",
    "ifd": "IFD",
    "illumination_type": "IlluminationType",
    "image": "Image",
    "image_ref": "ImageRef",
    "images": "Images",
    "imaging_environment": "ImagingEnvironment",
    "immersion": "Immersion",
    "immunocytochemistry": "Immunocytochemistry",
    "immunofluorescence": "Immunofluorescence",
    "incandescent": "Incandescent",
    "index": "Index",
    "institution": "Institution",
    "instrument": "Instrument",
    "instrument_ref": "InstrumentRef",
    "instruments": "Instruments",
    "integration": "Integration",
    "intensified_ccd": "IntensifiedCCD",
    "interleaved": "Interleaved",
    "inverse_frap": "InverseFRAP",
    "inverted": "Inverted",
    "ion_imaging": "IonImaging",
    "iris": "Iris",
    "italic": "Italic",
    "k": "K",
    "kr": "Kr",
    "kr_cl": "KrCl",
    "kr_fl": "KrFl",
    "lcm": "LCM",
    "label": "Label",
    "laser": "Laser",
    "laser_medium": "LaserMedium",
    "laser_scanning_confocal_microscopy": "LaserScanningConfocalMicroscopy",
    "last_name": "LastName",
    "leader": "Leader",
    "length": "Length",
    "lens_na": "LensNA",
    "lifetime_imaging": "LifetimeImaging",
    "light_emitting_diode": "LightEmittingDiode",
    "light_path": "LightPath",
    "light_source": "LightSource",
    "light_source_group": "LightSourceGroup",
    "light_source_settings": "LightSourceSettings",
    "line": "Line",
    "list_annotation": "ListAnnotation",
    "list_annotations": "ListAnnotations",
    "locked": "Locked",
    "long_annotation": "LongAnnotation",
    "long_annotations": "LongAnnotations",
    "long_pass": "LongPass",
    "lot_number": "LotNumber",
    "m": "M",
    "mime_type": "MIMEType",
    "manufacturer": "Manufacturer",
    "manufacturer_spec": "ManufacturerSpec",
    "map": "Map",
    "map_annotation": "MapAnnotation",
    "map_annotations": "MapAnnotations",
    "marker_end": "MarkerEnd",
    "marker_start": "MarkerStart",
    "mask": "Mask",
    "maximum_field_count": "MaximumFieldCount",
    "medium": "Medium",
    "metadata_file": "MetadataFile",
    "metadata_only": "MetadataOnly",
    "metal_vapor": "MetalVapor",
    "microbeam_manipulation": "MicrobeamManipulation",
    "microbeam_manipulation_ref": "MicrobeamManipulationRef",
    "microbeam_manipulations": "MicrobeamManipulations",
    "microscope": "Microscope",
    "middle_name": "MiddleName",
    "mode_locked": "ModeLocked",
    "model": "Model",
    "multi": "Multi",
    "multi_pass": "MultiPass",
    "multi_photon_microscopy": "MultiPhotonMicroscopy",
    "n": "N",
    "nd_filter": "NDFilter",
    "name": "Name",
    "namespace": "Namespace",
    "nd_glass": "NdGlass",
    "nd_yag": "NdYAG",
    "near_field_scanning_optical_microscopy": "NearFieldScanningOpticalMicroscopy",
    "neofluar": "Neofluar",
    "neutral_density": "NeutralDensity",
    "nominal_magnification": "NominalMagnification",
    "non_linear": "NonLinear",
    "non_zero": "NonZero",
    "normal": "Normal",
    "numeric_annotation": "NumericAnnotation",
    "ome": "OME",
    "objective": "Objective",
    "objective_settings": "ObjectiveSettings",
    "objectives": "Objectives",
    "oblique": "Oblique",
    "oblique_illumination": "ObliqueIllumination",
    "offset": "Offset",
    "oil": "Oil",
    "optical_trapping": "OpticalTrapping",
    "other": "Other",
    "palm": "PALM",
    "pgi_documentation": "PGIDocumentation",
    "pmt": "PMT",
    "phase": "Phase",
    "photoablation": "Photoablation",
    "photoactivation": "Photoactivation",
    "photobleaching": "Photobleaching",
    "photodiode": "Photodiode",
    "physical_size_x": "PhysicalSizeX",
    "physical_size_x_unit": "PhysicalSizeXUnit",
    "physical_size_y": "PhysicalSizeY",
    "physical_size_y_unit": "PhysicalSizeYUnit",
    "physical_size_z": "PhysicalSizeZ",
    "physical_size_z_unit": "PhysicalSizeZUnit",
    "pinhole_size": "PinholeSize",
    "pinhole_size_unit": "PinholeSizeUnit",
    "pixels": "Pixels",
    "plan_apo": "PlanApo",
    "plan_fluor": "PlanFluor",
    "plan_neofluar": "PlanNeofluar",
    "plane": "Plane",
    "plane_count": "PlaneCount",
    "planes": "Planes",
    "plate": "Plate",
    "plate_acquisition": "PlateAcquisition",
    "plate_acquisitions": "PlateAcquisitions",
    "plate_ref": "PlateRef",
    "plates": "Plates",
    "pockel_cell": "PockelCell",
    "pockel_cell_setting": "PockelCellSetting",
    "point": "Point",
    "points": "Points",
    "polarized_light": "PolarizedLight",
    "polygon": "Polygon",
    "polyline": "Polyline",
    "position_x": "PositionX",
    "position_x_unit": "PositionXUnit",
    "position_y": "PositionY",
    "position_y_unit": "PositionYUnit",
    "position_z": "PositionZ",
    "position_z_unit": "PositionZUnit",
    "power": "Power",
    "power_unit": "PowerUnit",
    "project": "Project",
    "project_ref": "ProjectRef",
    "projects": "Projects",
    "protocol_description": "ProtocolDescription",
    "protocol_identifier": "ProtocolIdentifier",
    "pulse": "Pulse",
    "pump": "Pump",
    "q_switched": "QSwitched",
    "roi": "ROI",
    "roi_ref": "ROIRef",
    "rois": "ROIs",
    "radius_x": "RadiusX",
    "radius_y": "RadiusY",
    "read_out_rate": "ReadOutRate",
    "read_out_rate_unit": "ReadOutRateUnit",
    "reagent": "Reagent",
    "reagent_identifier": "ReagentIdentifier",
    "reagent_ref": "ReagentRef",
    "reagent_set_description": "ReagentSetDescription",
    "reagent_set_identifier": "ReagentSetIdentifier",
    "reagents": "Reagents",
    "rectangle": "Rectangle",
    "reference": "Reference",
    "refractive_index": "RefractiveIndex",
    "repetition_rate": "RepetitionRate",
    "repetition_rate_unit": "RepetitionRateUnit",
    "repetitive": "Repetitive",
    "rhodamine6_g": "Rhodamine6G",
    "rights": "Rights",
    "rights_held": "RightsHeld",
    "rights_holder": "RightsHolder",
    "row": "Row",
    "row_naming_convention": "RowNamingConvention",
    "rows": "Rows",
    "ruby": "Ruby",
    "sha1": "SHA1",
    "spim": "SPIM",
    "sted": "STED",
    "storm": "STORM",
    "samples_per_pixel": "SamplesPerPixel",
    "screen": "Screen",
    "screens": "Screens",
    "second_harmonic_generation_imaging": "SecondHarmonicGenerationImaging",
    "semiconductor": "Semiconductor",
    "serial_number": "SerialNumber",
    "settings": "Settings",
    "shape": "Shape",
    "shape_group": "ShapeGroup",
    "short_pass": "ShortPass",
    "significant_bits": "SignificantBits",
    "single": "Single",
    "single_molecule_imaging": "SingleMoleculeImaging",
    "size": "Size",
    "size_c": "SizeC",
    "size_t": "SizeT",
    "size_x": "SizeX",
    "size_y": "SizeY",
    "size_z": "SizeZ",
    "slit_scan_confocal": "SlitScanConfocal",
    "solid_state": "SolidState",
    "spectral_imaging": "SpectralImaging",
    "spectroscopy": "Spectroscopy",
    "spinning_disk_confocal": "SpinningDiskConfocal",
    "stage_label": "StageLabel",
    "start_time": "StartTime",
    "status": "Status",
    "stroke_color": "StrokeColor",
    "stroke_dash_array": "StrokeDashArray",
    "stroke_width": "StrokeWidth",
    "stroke_width_unit": "StrokeWidthUnit",
    "structured_annotations": "StructuredAnnotations",
    "structured_illumination": "StructuredIllumination",
    "super_fluor": "SuperFluor",
    "swept_field_confocal": "SweptFieldConfocal",
    "tirf": "TIRF",
    "tag_annotation": "TagAnnotation",
    "tag_annotations": "TagAnnotations",
    "temperature": "Temperature",
    "temperature_unit": "TemperatureUnit",
    "term_annotation": "TermAnnotation",
    "term_annotations": "TermAnnotations",
    "text": "Text",
    "text_annotation": "TextAnnotation",
    "the_c": "TheC",
    "the_t": "TheT",
    "the_z": "TheZ",
    "ti_sapphire": "TiSapphire",
    "tiff_data": "TiffData",
    "tiff_data_blocks": "TiffDataBlocks",
    "time_increment": "TimeIncrement",
    "time_increment_unit": "TimeIncrementUnit",
    "time_lapse": "TimeLapse",
    "timepoint": "Timepoint",
    "timestamp_annotation": "TimestampAnnotation",
    "timestamp_annotations": "TimestampAnnotations",
    "total_internal_reflection": "TotalInternalReflection",
    "transform": "Transform",
    "transmittance": "Transmittance",
    "transmittance_range": "TransmittanceRange",
    "transmitted": "Transmitted",
    "tuneable": "Tuneable",
    "type": "Type",
    "type_annotation": "TypeAnnotation",
    "uuid": "UUID",
    "uv": "UV",
    "uncaging": "Uncaging",
    "union": "Union",
    "upright": "Upright",
    "user_name": "UserName",
    "value": "Value",
    "violet_corrected": "VioletCorrected",
    "voltage": "Voltage",
    "voltage_unit": "VoltageUnit",
    "water": "Water",
    "water_dipping": "WaterDipping",
    "wavelength": "Wavelength",
    "wavelength_unit": "WavelengthUnit",
    "well": "Well",
    "well_origin_x": "WellOriginX",
    "well_origin_x_unit": "WellOriginXUnit",
    "well_origin_y": "WellOriginY",
    "well_origin_y_unit": "WellOriginYUnit",
    "well_sample": "WellSample",
    "well_sample_ref": "WellSampleRef",
    "well_samples": "WellSamples",
    "wells": "Wells",
    "wide_field": "WideField",
    "width": "Width",
    "working_distance": "WorkingDistance",
    "working_distance_unit": "WorkingDistanceUnit",
    "x": "X",
    "x1": "X1",
    "x2": "X2",
    "xml_annotation": "XMLAnnotation",
    "xml_annotations": "XMLAnnotations",
    "x_unit": "XUnit",
    "xyctz": "XYCTZ",
    "xyczt": "XYCZT",
    "xytcz": "XYTCZ",
    "xytzc": "XYTZC",
    "xyzct": "XYZCT",
    "xyztc": "XYZTC",
    "xe": "Xe",
    "xe_br": "XeBr",
    "xe_cl": "XeCl",
    "xe_fl": "XeFl",
    "y": "Y",
    "y1": "Y1",
    "y2": "Y2",
    "y_unit": "YUnit",
    "z": "Z",
    "z_unit": "ZUnit",
    "zoom": "Zoom",
    "bzip2": "bzip2",
    "cursive": "cursive",
    "fantasy": "fantasy",
    "href": "href",
    "letter": "letter",
    "monospace": "monospace",
    "none": "none",
    "number": "number",
    "sans-serif": "sans-serif",
    "serif": "serif",
    "zlib": "zlib",
}

_lists = {
    "Annotation": {"AnnotationRef"},
    "LightSource": {"AnnotationRef"},
    "Map": {"M"},
    "Shape": {"AnnotationRef"},
    "OME": {
        "Dataset",
        "Instrument",
        "Experimenter",
        "Experiment",
        "Folder",
        "ExperimenterGroup",
        "Project",
        "ROI",
        "Image",
        "Screen",
        "Plate",
    },
    "Image": {"MicrobeamManipulationRef", "ROIRef", "AnnotationRef"},
    "Pixels": {"TiffData", "Channel", "BinData", "Plane"},
    "Plane": {"AnnotationRef"},
    "Channel": {"AnnotationRef"},
    "MicrobeamManipulation": {"LightSourceSettings", "ROIRef", "Type"},
    "Instrument": {
        "Dichroic",
        "AnnotationRef",
        "Objective",
        "Detector",
        "Filter",
        "FilterSet",
    },
    "Project": {"AnnotationRef", "DatasetRef"},
    "ExperimenterGroup": {"Leader", "ExperimenterRef", "AnnotationRef"},
    "Dataset": {"ImageRef", "AnnotationRef"},
    "Experiment": {"MicrobeamManipulation", "Type"},
    "Experimenter": {"AnnotationRef"},
    "Folder": {"ImageRef", "ROIRef", "AnnotationRef", "FolderRef"},
    "Objective": {"AnnotationRef"},
    "Detector": {"AnnotationRef"},
    "FilterSet": {"ExcitationFilterRef", "EmissionFilterRef"},
    "Filter": {"AnnotationRef"},
    "Dichroic": {"AnnotationRef"},
    "LightPath": {"ExcitationFilterRef", "EmissionFilterRef", "AnnotationRef"},
    "ROI": {"AnnotationRef"},
    "Plate": {"AnnotationRef", "PlateAcquisition", "Well"},
    "Reagent": {"AnnotationRef"},
    "Screen": {"AnnotationRef", "PlateRef", "Reagent"},
    "PlateAcquisition": {"AnnotationRef", "WellSampleRef"},
    "Well": {"WellSample", "AnnotationRef"},
}



from __future__ import annotations

import os.path
from collections import defaultdict
from datetime import datetime
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import IO, Any, Union
from xml.etree import ElementTree

import xmlschema
from xmlschema import ElementData, XMLSchemaParseError
from xmlschema.converters import XMLSchemaConverter
from xmlschema.exceptions import XMLSchemaValueError

from ome_types._base_type import OMEType

from . import util
from ._constants import NS_OME, NS_XSI, OME_2016_06_XSD, SCHEMA_LOC_OME, URI_OME
from .model import (
    OME,
    XMLAnnotation,
    _camel_to_snake,
    _plural_to_singular,
    _singular_to_plural,
    _snake_to_camel,
    simple_types,
)

__cache__: dict[str, xmlschema.XMLSchema] = {}
_XMLSCHEMA_VERSION: tuple[int, ...] = tuple(
    int(v) if v.isnumeric() else v for v in xmlschema.__version__.split(".")
)

XMLSourceType = Union[str, bytes, Path, IO[str], IO[bytes]]


@lru_cache(maxsize=8)
def _build_schema(ns: str, uri: str | None = None) -> xmlschema.XMLSchema:
    """Return Schema object for a url.

    For the special case of retrieving the 2016-06 OME Schema, use local file.
    """
    if ns == URI_OME:
        schema = xmlschema.XMLSchema(OME_2016_06_XSD)
        # FIXME Hack to work around xmlschema poor support for keyrefs to
        # substitution groups
        ls_sgs = schema.maps.substitution_groups[f"{NS_OME}LightSourceGroup"]
        ls_id_maps = schema.maps.identities[f"{NS_OME}LightSourceIDKey"]
        ls_id_maps.elements = {e: None for e in ls_sgs}
    else:
        schema = xmlschema.XMLSchema(uri)
    return schema


def get_schema(source: xmlschema.XMLResource | XMLSourceType) -> xmlschema.XMLSchema:
    """Fetch an XMLSchema object given XML source.

    Parameters
    ----------
    source : XMLResource or str
        can be an :class:`xmlschema.XMLResource` instance, a file-like object, a path
        to a file or an URI of a resource or an Element instance or an ElementTree
        instance or a string containing the XML data.

    Returns
    -------
    xmlschema.XMLSchema
        An XMLSchema object for the source
    """
    if not isinstance(source, xmlschema.XMLResource):
        source = xmlschema.XMLResource(source)

    for ns, uri in source.get_locations():
        try:
            return _build_schema(ns, uri)
        except XMLSchemaParseError:
            pass
    raise XMLSchemaValueError(f"Could not find a schema for XML resource {source!r}.")


def validate(xml: XMLSourceType, schema: xmlschema.XMLSchema | None = None) -> None:
    schema = schema or get_schema(xml)
    schema.validate(xml)


class OMEConverter(XMLSchemaConverter):
    def __init__(
        self, namespaces: dict[str, Any] | None = None, **kwargs: dict[Any, Any]
    ):
        self._ome_ns = ""
        super().__init__(namespaces, attr_prefix="")
        for name, uri in self._namespaces.items():
            if uri == URI_OME:
                self._ome_ns = name

    def map_qname(self, qname: str) -> str:
        name = super().map_qname(qname)
        if name.lower().startswith(self._ome_ns):
            name = name[len(self._ome_ns) :].lstrip(":")
        return _camel_to_snake.get(name, name)

    def element_decode(self, data, xsd_element, xsd_type=None, level=0):  # type: ignore
        """Convert a decoded element data to a data structure."""
        result = super().element_decode(data, xsd_element, xsd_type, level)
        if isinstance(result, dict) and "$" in result:
            result["value"] = result.pop("$")
        # FIXME: Work out a better way to deal with concrete extensions of
        # abstract types.
        if xsd_element.local_name == "MetadataOnly":
            result = True
        elif xsd_element.local_name == "BinData":
            if result["length"] == 0 and "value" not in result:
                result["value"] = ""
        elif xsd_element.local_name == "StructuredAnnotations":
            annotations = []
            for _type in (
                "boolean_annotation",
                "comment_annotation",
                "double_annotation",
                "file_annotation",
                "list_annotation",
                "long_annotation",
                "map_annotation",
                "tag_annotation",
                "term_annotation",
                "timestamp_annotation",
                "xml_annotation",
            ):
                if _type in result:
                    values = result.pop(_type)
                    for v in values:
                        v["_type"] = _type
                        # Normalize empty element to zero-length string.
                        if "value" in v and v["value"] is None:
                            v["value"] = ""
                    annotations.extend(values)
            result = annotations
        if isinstance(result, dict):
            for name in list(result.keys()):
                plural = _singular_to_plural.get((xsd_element.local_name, name), None)
                if plural:
                    value = result.pop(name)
                    if not isinstance(value, list):
                        raise TypeError("expected list for plural attr")
                    result[plural] = value
        return result

    def element_encode(
        self, obj: Any, xsd_element: xmlschema.XsdElement, level: int = 0
    ) -> ElementData:
        tag = xsd_element.qualified_name
        if not isinstance(obj, OMEType):
            if isinstance(obj, datetime):
                return ElementData(
                    tag, obj.isoformat().replace("+00:00", "Z"), None, {}
                )
            elif isinstance(obj, ElementTree.Element):
                # ElementData can't represent mixed content, so we'll leave this
                # element empty and fix it up after encoding is complete.
                return ElementData(tag, None, None, {})
            elif xsd_element.type.simple_type is not None:
                return ElementData(tag, obj, None, {})
            elif xsd_element.local_name == "MetadataOnly":
                return ElementData(tag, None, None, {})
            elif xsd_element.local_name in {"Union", "StructuredAnnotations"}:
                names = [type(v).__name__ for v in obj]
                content = [(f"{NS_OME}{n}", v) for n, v in zip(names, obj)]
                return ElementData(tag, None, content, {})
            else:
                raise NotImplementedError(
                    "Encountered a combination of schema element and data type"
                    " that is not yet supported. Please submit a bug report with"
                    " the information below:"
                    f"\n    element: {xsd_element}\n    data type: {type(obj)}"
                )
        text = None
        content = []
        attributes = {}
        # FIXME Can we simplify this?
        tag_index = defaultdict(
            lambda: -1,
            ((_camel_to_snake[e.local_name], i) for i, e in enumerate(xsd_element)),
        )
        _fields = obj.__fields__.values()
        for field in sorted(
            _fields,
            key=lambda f: tag_index[_plural_to_singular.get(f.name, f.name)],
        ):
            name = field.name
            if name.endswith("_"):
                continue
            default = (
                field.default_factory() if field.default_factory else field.default
            )
            value = getattr(obj, name)
            if value == default or name == "metadata_only" and not value:
                continue
            if isinstance(value, simple_types.Color):
                value = value.as_int32()
            name = _plural_to_singular.get(name, name)
            name = _snake_to_camel.get(name, name)
            if name in xsd_element.attributes:
                if isinstance(value, list):
                    value = [getattr(i, "value", i) for i in value]
                elif isinstance(value, Enum):
                    value = value.value
                elif isinstance(value, datetime):
                    value = value.isoformat().replace("+00:00", "Z")
                attributes[name] = value
            elif name == "Value" and xsd_element.local_name in {"BinData", "UUID", "M"}:
                text = value
            else:
                if not isinstance(value, list) or name in {
                    "Union",
                    "StructuredAnnotations",
                }:
                    value = [value]
                if name == "LightSourceGroup":
                    names = [type(v).__name__ for v in value]
                else:
                    names = [name] * len(value)
                content.extend([(f"{NS_OME}{n}", v) for n, v in zip(names, value)])
        return ElementData(tag, text, content, attributes)


def xmlschema2dict(
    xml: str,
    schema: xmlschema.XMLSchema | None = None,
    converter: XMLSchemaConverter = OMEConverter,
    validate: bool = False,
    **kwargs: Any,
) -> dict[str, Any]:
    if isinstance(xml, bytes):
        xml = xml.decode("utf-8")

    schema = schema or get_schema(xml)

    if _XMLSCHEMA_VERSION >= (2,):
        kwargs["validation"] = "strict" if validate else "lax"

    result = xmlschema.to_dict(xml, schema=schema, converter=converter, **kwargs)

    if _XMLSCHEMA_VERSION >= (2,) and not validate:
        result, _ = result
    # xmlschema doesn't provide usable access to mixed XML content, so we'll
    # fill the XMLAnnotation value attributes ourselves by re-parsing the XML
    # with ElementTree and using the Element objects as the values.
    tree = None
    for annotation in result.get("structured_annotations", []):
        if annotation["_type"] == "xml_annotation":
            if tree is None:
                from io import StringIO

                # determine if we're dealing with a raw XML string or a filepath
                # very long XML strings will raise ValueError on Windows.
                try:
                    _xml = xml if os.path.exists(xml) else StringIO(xml)
                except ValueError:
                    _xml = StringIO(xml)

                tree = ElementTree.parse(_xml)  # type: ignore  # noqa: S314
            aid = annotation["id"]
            elt = tree.find(f".//{NS_OME}XMLAnnotation[@ID='{aid}']/{NS_OME}Value")
            annotation["value"] = elt
    return result


def to_xml_element(ome: OME) -> ElementTree.Element:
    schema = _build_schema(URI_OME)
    root = schema.encode(
        ome, path=f"/{NS_OME}OME", converter=OMEConverter, use_defaults=False
    )
    # Patch up the XML element tree with Element objects from XMLAnnotations to
    # work around xmlschema's lack of support for mixed content.
    for oid, obj in util.collect_ids(ome).items():
        if isinstance(obj, XMLAnnotation):
            elt = root.find(f".//{NS_OME}XMLAnnotation[@ID='{oid}']/{NS_OME}Value")
            elt.extend(list(obj.value))
    root.attrib[f"{NS_XSI}schemaLocation"] = SCHEMA_LOC_OME
    return root


def to_xml(ome: OME, **kwargs: Any) -> str:
    """
    Dump an OME object to string.

    Parameters
    ----------
    ome: OME
        OME object to dump.
    **kwargs
        Extra kwargs to pass to ElementTree.tostring.

    Returns
    -------
    ome_string: str
        The XML string of the OME object.
    """
    root = to_xml_element(ome)
    ElementTree.register_namespace("", URI_OME)
    kwargs.setdefault("encoding", "unicode")
    kwargs.setdefault("method", "xml")
    return ElementTree.tostring(root, **kwargs)



from ome_types._base_type import OMEType


class AffineTransform(OMEType):
    """A matrix used to transform the shape.

    ⎡ A00, A01, A02 ⎤ ⎢ A10, A11, A12 ⎥ ⎣ 0,   0,   1   ⎦

    Parameters
    ----------
    a00 : float
    a01 : float
    a02 : float
    a10 : float
    a11 : float
    a12 : float
    """

    a00: float
    a01: float
    a02: float
    a10: float
    a11: float
    a12: float



from enum import Enum
from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .experimenter_ref import ExperimenterRef
from .light_source_settings import LightSourceSettings
from .roi_ref import ROIRef
from .simple_types import MicrobeamManipulationID


class Type(Enum):
    """The type of manipulation performed."""

    FLIP = "FLIP"
    FRAP = "FRAP"
    INVERSE_FRAP = "InverseFRAP"
    OPTICAL_TRAPPING = "OpticalTrapping"
    OTHER = "Other"
    PHOTOABLATION = "Photoablation"
    PHOTOACTIVATION = "Photoactivation"
    UNCAGING = "Uncaging"


class MicrobeamManipulation(OMEType):
    """Defines a microbeam operation type and the region of the image it was applied
    to.

    The LightSourceRef element is a reference to a LightSource specified in the
    Instrument element which was used for a technique other than illumination for
    the purpose of imaging. For example, a laser used for photobleaching.

    Parameters
    ----------
    experimenter_ref : ExperimenterRef
    id : MicrobeamManipulationID
    roi_ref : ROIRef
    description : str, optional
        A description for the Microbeam Manipulation.
    light_source_settings : LightSourceSettings, optional
    type : Type, optional
        The type of manipulation performed.
    """

    experimenter_ref: ExperimenterRef
    id: MicrobeamManipulationID
    roi_ref: List[ROIRef]
    description: Optional[str] = None
    light_source_settings: List[LightSourceSettings] = Field(default_factory=list)
    type: List[Type] = Field(default_factory=list)



from typing import TYPE_CHECKING, Any, Optional
from weakref import ReferenceType

from ome_types._base_type import OMEType

from .simple_types import LSID


class Reference(OMEType):
    """Reference is an empty complex type that is contained and extended by all the
    `*Ref` elements and also the Settings Complex Type Each `*Ref` element defines
    an attribute named ID of simple type `*ID` and no other information Each
    simple type `*ID` is restricted to the base type LSID with an appropriate
    pattern
    """

    if TYPE_CHECKING:
        _ref: Optional["ReferenceType[OMEType]"]

    id: LSID
    _ref = None

    @property
    def ref(self) -> Any:
        if self._ref is None:
            raise ValueError("references not yet resolved on root OME object")
        return self._ref()



import pytest
from pint import DimensionalityError
from pydantic import ValidationError

from ome_types._units import ureg
from ome_types.model import Channel, Laser, Plane, simple_types


def test_quantity_math():
    """Validate math on quantities with different but compatible units."""
    channel = Channel(
        excitation_wavelength=475,
        excitation_wavelength_unit="nm",
        emission_wavelength=530000,
        emission_wavelength_unit="pm",
    )
    shift = (
        channel.emission_wavelength_quantity - channel.excitation_wavelength_quantity
    )
    # Compare to a tolerance due to Pint internal factor representation.
    assert abs(shift.to("nm").m - 55) < 1e-12


def test_invalid_unit():
    """Ensure incompatible units in constructor raises ValidationError."""
    with pytest.raises(ValidationError):
        Channel(
            excitation_wavelength=475,
            excitation_wavelength_unit="kg",
        )


def test_dimensionality_error():
    """Ensure math on incompatible units raises DimensionalityError."""
    laser = Laser(
        id="LightSource:1",
        repetition_rate=10,
        repetition_rate_unit="MHz",
        wavelength=640,
    )
    with pytest.raises(DimensionalityError):
        laser.repetition_rate_quantity + laser.wavelength_quantity


def test_reference_frame():
    """Validate reference_frame behavior."""
    plane = Plane(
        the_c=0,
        the_t=0,
        the_z=0,
        position_x=1,
        position_x_unit="reference frame",
        position_y=2,
        position_y_unit="mm",
    )
    # Verify two different ways that reference_frame and length are incompatible.
    with pytest.raises(DimensionalityError):
        plane.position_x_quantity + plane.position_y_quantity
    product = plane.position_x_quantity * plane.position_y_quantity
    assert not product.check("[area]")
    # Verify that we can obtain a usable length if we know the conversion factor.
    conversion_factor = ureg.Quantity(1, "micron/reference_frame")
    position_x = plane.position_x_quantity * conversion_factor
    assert position_x.check("[length]")


def test_all_units():
    """Test that all Unit* enums are in the registry."""
    for t in dir(simple_types):
        if not t.startswith("Unit"):
            continue
        e = getattr(simple_types, t)
        for v in e.__members__.values():
            assert v.value.replace(" ", "_") in ureg



from typing import Optional

from ome_types._base_type import OMEType


class ManufacturerSpec(OMEType):
    """This is the base from which many microscope components are extended.

    E.g Objective, Filter etc. Provides attributes for recording common properties
    of these components such as Manufacturer name, Model etc, all of which are
    optional.

    Parameters
    ----------
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    serial_number : str, optional
        The serial number of the component.
    """

    lot_number: Optional[str] = None
    manufacturer: Optional[str] = None
    model: Optional[str] = None
    serial_number: Optional[str] = None



from enum import Enum
from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .experimenter_ref import ExperimenterRef
from .microbeam_manipulation import MicrobeamManipulation
from .simple_types import ExperimentID


class Type(Enum):
    """A term to describe the type of experiment."""

    COLOCALIZATION = "Colocalization"
    ELECTROPHYSIOLOGY = "Electrophysiology"
    FISH = "FISH"
    FLUORESCENCE_LIFETIME = "FluorescenceLifetime"
    FOUR_D_PLUS = "FourDPlus"
    FP = "FP"
    FRET = "FRET"
    IMMUNOCYTOCHEMISTRY = "Immunocytochemistry"
    IMMUNOFLUORESCENCE = "Immunofluorescence"
    ION_IMAGING = "IonImaging"
    OTHER = "Other"
    PGI_DOCUMENTATION = "PGIDocumentation"
    PHOTOBLEACHING = "Photobleaching"
    SCREEN = "Screen"
    SPECTRAL_IMAGING = "SpectralImaging"
    SPIM = "SPIM"
    TIME_LAPSE = "TimeLapse"


class Experiment(OMEType):
    """This element describes the type of experiment.

    The required Type attribute must contain one or more entries from the
    following list: FP FRET Time-lapse 4-D+ Screen Immunocytochemistry FISH
    Electrophysiology  Ion-Imaging Colocalization PGI/Documentation FRAP
    Photoablation Optical-Trapping Photoactivation Fluorescence-Lifetime Spectral-
    Imaging Other FP refers to fluorescent proteins, PGI/Documentation is not a
    'data' image. The optional Description element may contain free text to
    further describe the experiment.

    Parameters
    ----------
    id : ExperimentID
    description : str, optional
        A description for the experiment.
    experimenter_ref : ExperimenterRef, optional
        This is a link to the Experimenter who conducted the experiment
    microbeam_manipulations : MicrobeamManipulation, optional
    type : Type, optional
        A term to describe the type of experiment.
    """

    id: ExperimentID
    description: Optional[str] = None
    experimenter_ref: Optional[ExperimenterRef] = None
    microbeam_manipulations: List[MicrobeamManipulation] = Field(default_factory=list)
    type: List[Type] = Field(default_factory=list)



from enum import Enum
from typing import Optional

from typing_extensions import Literal

from ome_types._base_type import OMEType

from .light_source import LightSource
from .pump import Pump
from .simple_types import PositiveFloat, PositiveInt, UnitsFrequency, UnitsLength


class Type(Enum):
    """Type is the general category of laser."""

    DYE = "Dye"
    EXCIMER = "Excimer"
    FREE_ELECTRON = "FreeElectron"
    GAS = "Gas"
    METAL_VAPOR = "MetalVapor"
    OTHER = "Other"
    SEMICONDUCTOR = "Semiconductor"
    SOLID_STATE = "SolidState"


class LaserMedium(Enum):
    """The Medium attribute specifies the actual lasing medium for a given laser
    type.
    """

    AG = "Ag"
    ALEXANDRITE = "Alexandrite"
    AR = "Ar"
    AR_CL = "ArCl"
    AR_FL = "ArFl"
    CO = "CO"
    CO2 = "CO2"
    COUMARIN_C30 = "CoumarinC30"
    CU = "Cu"
    E_MINUS = "EMinus"
    ER_GLASS = "ErGlass"
    ER_YAG = "ErYAG"
    GA_AL_AS = "GaAlAs"
    GA_AS = "GaAs"
    H2_O = "H2O"
    H_FL = "HFl"
    HE_CD = "HeCd"
    HE_NE = "HeNe"
    HO_YAG = "HoYAG"
    HO_YLF = "HoYLF"
    KR = "Kr"
    KR_CL = "KrCl"
    KR_FL = "KrFl"
    N = "N"
    ND_GLASS = "NdGlass"
    ND_YAG = "NdYAG"
    OTHER = "Other"
    RHODAMINE6_G = "Rhodamine6G"
    RUBY = "Ruby"
    TI_SAPPHIRE = "TiSapphire"
    XE = "Xe"
    XE_BR = "XeBr"
    XE_CL = "XeCl"
    XE_FL = "XeFl"


class Pulse(Enum):
    """The Pulse mode of the laser."""

    CW = "CW"
    MODE_LOCKED = "ModeLocked"
    OTHER = "Other"
    Q_SWITCHED = "QSwitched"
    REPETITIVE = "Repetitive"
    SINGLE = "Single"


class Laser(LightSource, OMEType):
    """Laser types are specified using two attributes - the Type and the LaserMedium.

    Parameters
    ----------
    id : LightSourceID
        A LightSource ID must be specified for each light source, and the
        individual light sources can be referred to by their LightSource IDs
        (eg from Channel).
    annotation_ref : AnnotationRef, optional
    frequency_multiplication : PositiveInt, optional
        FrequencyMultiplication that may be specified.
    laser_medium : LaserMedium, optional
        The Medium attribute specifies the actual lasing medium for a given
        laser type.
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    pockel_cell : bool, optional
        If true the laser has a PockelCell to rotate the polarization of the
        beam.
    power : float, optional
        The light-source power. Units are set by PowerUnit.
    power_unit : UnitsPower, optional
        The units of the Power - default:milliwatts.
    pulse : Pulse, optional
        The Pulse mode of the laser.
    pump : Pump, optional
        The Laser element may contain a Pump sub-element which refers to a
        LightSource used as a laser pump.
    repetition_rate : float, optional
        The is the rate in Hz at which the laser pulses if the Pulse type is
        'Repetitive'. hertz Units are set by RepetitionRateUnit.
    repetition_rate_unit : UnitsFrequency, optional
        The units of the RepetitionRate - default:hertz.
    serial_number : str, optional
        The serial number of the component.
    tuneable : bool, optional
        Whether or not the laser is Tuneable
    type : Type, optional
        Type is the general category of laser.
    wavelength : PositiveFloat, optional
        The Wavelength of the laser. Units are set by WavelengthUnit.
    wavelength_unit : UnitsLength, optional
        The units of the Wavelength - default:nanometres.
    """

    kind: Literal["laser"] = "laser"
    frequency_multiplication: Optional[PositiveInt] = None
    laser_medium: Optional[LaserMedium] = None
    pockel_cell: Optional[bool] = None
    pulse: Optional[Pulse] = None
    pump: Optional[Pump] = None
    repetition_rate: Optional[float] = None
    repetition_rate_unit: Optional[UnitsFrequency] = UnitsFrequency("Hz")
    tuneable: Optional[bool] = None
    type: Optional[Type] = None
    wavelength: Optional[PositiveFloat] = None
    wavelength_unit: Optional[UnitsLength] = UnitsLength("nm")



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .simple_types import ReagentID


class Reagent(OMEType):
    """Reagent is used to describe a chemical or some other physical experimental
    parameter.

    Parameters
    ----------
    id : ReagentID
    annotation_ref : AnnotationRef, optional
    description : str, optional
        A long description for the reagent.
    name : str, optional
        A short name for the reagent
    reagent_identifier : str, optional
        This is a reference to an external (to OME) representation of the
        Reagent. It serves as a foreign key into an external database. - It is
        sometimes referred to as ExternalIdentifier.
    """

    id: ReagentID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    description: Optional[str] = None
    name: Optional[str] = None
    reagent_identifier: Optional[str] = None



from typing import Optional

from ome_types._base_type import OMEType

from .simple_types import NonNegativeFloat, PercentFraction, PositiveFloat, UnitsLength


class TransmittanceRange(OMEType):
    """This records the range of wavelengths that are transmitted by the filter.

    It also records the maximum amount of light transmitted.

    Parameters
    ----------
    cut_in : PositiveFloat, optional
        CutIn is the wavelength below which there is less than 50%
        transmittance for a filter. Units are set by CutInUnit.
    cut_in_tolerance : NonNegativeFloat, optional
        CutInTolerance. Units are set by CutInToleranceUnit.
    cut_in_tolerance_unit : UnitsLength, optional
        The units of the CutInTolerance - default:nanometres.
    cut_in_unit : UnitsLength, optional
        The units of the CutIn - default:nanometres.
    cut_out : PositiveFloat, optional
        CutOut is the wavelength above which there is less than 50%
        transmittance for a filter. Units are set by CutOutUnit.
    cut_out_tolerance : NonNegativeFloat, optional
        CutOutTolerance. Units are set by CutOutToleranceUnit.
    cut_out_tolerance_unit : UnitsLength, optional
        The units of the CutOutTolerance - default:nanometres.
    cut_out_unit : UnitsLength, optional
        The units of the CutOut - default:nanometres.
    transmittance : PercentFraction, optional
        The amount of light the filter transmits at a maximum A fraction, as a
        value from 0.0 to 1.0.
    """

    cut_in: Optional[PositiveFloat] = None
    cut_in_tolerance: Optional[NonNegativeFloat] = None
    cut_in_tolerance_unit: Optional[UnitsLength] = UnitsLength("nm")
    cut_in_unit: Optional[UnitsLength] = UnitsLength("nm")
    cut_out: Optional[PositiveFloat] = None
    cut_out_tolerance: Optional[NonNegativeFloat] = None
    cut_out_tolerance_unit: Optional[UnitsLength] = UnitsLength("nm")
    cut_out_unit: Optional[UnitsLength] = UnitsLength("nm")
    transmittance: Optional[PercentFraction] = None



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .folder_ref import FolderRef
from .image_ref import ImageRef
from .roi_ref import ROIRef
from .simple_types import FolderID


class Folder(OMEType):
    """An element specifying a possibly heterogeneous collection of data.

    Folders may contain Folders so that data may be organized within a tree of
    Folders. Data may be in multiple Folders but a Folder may not be in more than
    one other Folder.

    Parameters
    ----------
    id : FolderID
    annotation_ref : AnnotationRef, optional
    description : str, optional
        A description for the folder.
    folder_ref : FolderRef, optional
    image_ref : ImageRef, optional
    name : str, optional
        A name for the folder that is suitable for presentation to the user.
    roi_ref : ROIRef, optional
    """

    id: FolderID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    description: Optional[str] = None
    folder_ref: List[FolderRef] = Field(default_factory=list)
    image_ref: List[ImageRef] = Field(default_factory=list)
    name: Optional[str] = None
    roi_ref: List[ROIRef] = Field(default_factory=list)



from enum import Enum
from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .detector_settings import DetectorSettings
from .filter_set_ref import FilterSetRef
from .light_path import LightPath
from .light_source_settings import LightSourceSettings
from .simple_types import ChannelID, Color, PositiveFloat, PositiveInt, UnitsLength


class IlluminationType(Enum):
    """The method of illumination used to capture the channel."""

    EPIFLUORESCENCE = "Epifluorescence"
    NON_LINEAR = "NonLinear"
    OBLIQUE = "Oblique"
    OTHER = "Other"
    TRANSMITTED = "Transmitted"


class AcquisitionMode(Enum):
    """AcquisitionMode describes the type of microscopy performed for each channel."""

    BRIGHT_FIELD = "BrightField"
    FLUORESCENCE_CORRELATION_SPECTROSCOPY = "FluorescenceCorrelationSpectroscopy"
    FLUORESCENCE_LIFETIME = "FluorescenceLifetime"
    FSM = "FSM"
    LASER_SCANNING_CONFOCAL_MICROSCOPY = "LaserScanningConfocalMicroscopy"
    LCM = "LCM"
    MULTI_PHOTON_MICROSCOPY = "MultiPhotonMicroscopy"
    NEAR_FIELD_SCANNING_OPTICAL_MICROSCOPY = "NearFieldScanningOpticalMicroscopy"
    OTHER = "Other"
    PALM = "PALM"
    SECOND_HARMONIC_GENERATION_IMAGING = "SecondHarmonicGenerationImaging"
    SINGLE_MOLECULE_IMAGING = "SingleMoleculeImaging"
    SLIT_SCAN_CONFOCAL = "SlitScanConfocal"
    SPECTRAL_IMAGING = "SpectralImaging"
    SPIM = "SPIM"
    SPINNING_DISK_CONFOCAL = "SpinningDiskConfocal"
    STED = "STED"
    STORM = "STORM"
    STRUCTURED_ILLUMINATION = "StructuredIllumination"
    SWEPT_FIELD_CONFOCAL = "SweptFieldConfocal"
    TIRF = "TIRF"
    TOTAL_INTERNAL_REFLECTION = "TotalInternalReflection"
    WIDE_FIELD = "WideField"


class ContrastMethod(Enum):
    """ContrastMethod describes the technique used to achieve contrast for each
    channel.
    """

    BRIGHTFIELD = "Brightfield"
    DARKFIELD = "Darkfield"
    DIC = "DIC"
    FLUORESCENCE = "Fluorescence"
    HOFFMAN_MODULATION = "HoffmanModulation"
    OBLIQUE_ILLUMINATION = "ObliqueIllumination"
    OTHER = "Other"
    PHASE = "Phase"
    POLARIZED_LIGHT = "PolarizedLight"


class Channel(OMEType):
    """There must be one per channel in the Image, even for a single-plane image.

    And information about how each of them was acquired is stored in the various
    optional `*Ref` elements.  Each Logical Channel is composed of one or more
    ChannelComponents.  For example, an entire spectrum in an FTIR experiment may
    be stored in a single Logical Channel with each discrete wavenumber of the
    spectrum constituting a ChannelComponent of the FTIR Logical Channel.  An RGB
    image where the Red, Green and Blue components do not reflect discrete probes
    but are instead the output of a color camera would be treated similarly - one
    Logical channel with three ChannelComponents in this case. The total number of
    ChannelComponents for a set of pixels must equal SizeC. The IlluminationType
    attribute is a string enumeration which may be set to 'Transmitted',
    'Epifluorescence', 'Oblique', or 'NonLinear'. The user interface logic for
    labeling a given channel for the user should use the first existing attribute
    in the following sequence: Name -> Fluor -> EmissionWavelength ->
    ChannelComponent/Index.

    Parameters
    ----------
    id : ChannelID
    acquisition_mode : AcquisitionMode, optional
        AcquisitionMode describes the type of microscopy performed for each
        channel
    annotation_ref : AnnotationRef, optional
    color : Color, optional
        A color used to render this channel - encoded as RGBA The default
        value "-1" is #FFFFFFFF so solid white (it is a signed 32 bit value)
        NOTE: Prior to the 2012-06 schema the default value was incorrect and
        produced a transparent red not solid white.
    contrast_method : ContrastMethod, optional
        ContrastMethod describes the technique used to achieve contrast for
        each channel
    detector_settings : DetectorSettings, optional
    emission_wavelength : PositiveFloat, optional
        Wavelength of emission for a particular channel. Units are set by
        EmissionWavelengthUnit.
    emission_wavelength_unit : UnitsLength, optional
        The units of the wavelength of emission - default:nanometres.
    excitation_wavelength : PositiveFloat, optional
        Wavelength of excitation for a particular channel. Units are set by
        ExcitationWavelengthUnit.
    excitation_wavelength_unit : UnitsLength, optional
        The units of the wavelength of excitation - default:nanometres.
    filter_set_ref : FilterSetRef, optional
    fluor : str, optional
        The Fluor attribute is used for fluorescence images. This is the name
        of the fluorophore used to produce this channel
    illumination_type : IlluminationType, optional
        The method of illumination used to capture the channel.
    light_path : LightPath, optional
    light_source_settings : LightSourceSettings, optional
    name : str, optional
        A name for the channel that is suitable for presentation to the user.
    nd_filter : float, optional
        The NDfilter attribute is used to specify the combined effect of any
        neutral density filters used. The amount of light the filter transmits
        at a maximum A fraction, as a value from 0.0 to 1.0.  NOTE: This was
        formerly described as "units optical density expressed as a
        PercentFraction". This was how the field had been described in the
        schema from the beginning but all the use of it has been in the
        opposite direction, i.e. as a amount transmitted, not the amount
        blocked. This change has been made to make the model reflect this
        usage.
    pinhole_size : float, optional
        The optional PinholeSize attribute allows specifying adjustable pin
        hole diameters for confocal microscopes. Units are set by
        PinholeSizeUnit.
    pinhole_size_unit : UnitsLength, optional
        The units of the pin hole diameter for confocal microscopes -
        default:microns.
    pockel_cell_setting : int, optional
        The PockelCellSetting used for this channel. This is the amount the
        polarization of the beam is rotated by.
    samples_per_pixel : PositiveInt, optional
        The number of samples the detector takes to form each pixel value.
        Note: This is not the same as "Frame Averaging" - see Integration in
        DetectorSettings
    """

    id: ChannelID
    acquisition_mode: Optional[AcquisitionMode] = None
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    color: Optional[Color] = Color("white")
    contrast_method: Optional[ContrastMethod] = None
    detector_settings: Optional[DetectorSettings] = None
    emission_wavelength: Optional[PositiveFloat] = None
    emission_wavelength_unit: Optional[UnitsLength] = UnitsLength("nm")
    excitation_wavelength: Optional[PositiveFloat] = None
    excitation_wavelength_unit: Optional[UnitsLength] = UnitsLength("nm")
    filter_set_ref: Optional[FilterSetRef] = None
    fluor: Optional[str] = None
    illumination_type: Optional[IlluminationType] = None
    light_path: Optional[LightPath] = None
    light_source_settings: Optional[LightSourceSettings] = None
    name: Optional[str] = None
    nd_filter: Optional[float] = None
    pinhole_size: Optional[float] = None
    pinhole_size_unit: Optional[UnitsLength] = UnitsLength("µm")
    pockel_cell_setting: Optional[int] = None
    samples_per_pixel: Optional[PositiveInt] = None



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .reagent import Reagent
from .reference import Reference
from .simple_types import PlateID, ScreenID


class PlateRef(Reference, OMEType):
    """The PlateRef element is a reference to a Plate element.

    Screen elements may have one or more PlateRef elements to define the plates
    that are part of the screen. Plates may belong to more than one screen.

    Parameters
    ----------
    id : PlateID
    """

    id: PlateID


class Screen(OMEType):
    """The Screen element is a grouping for Plates.

    The required attribute is the Screen's Name and ID - both must be unique
    within the document. The Screen element may contain an ExternalRef attribute
    that refers to an external database. A description of the screen may be
    specified in the Description element. Screens may contain overlapping sets of
    Plates i.e.      Screens and Plates have a many-to-many relationship. Plates
    contain one or more ScreenRef elements to specify what screens they belong to.

    Parameters
    ----------
    id : ScreenID
    annotation_ref : AnnotationRef, optional
    description : str, optional
        A description for the screen.
    name : str, optional
    plate_ref : PlateRef, optional
        The PlateRef element is a reference to a Plate element. Screen
        elements may have one or more PlateRef elements to define the plates
        that are part of the screen. Plates may belong to more than one
        screen.
    protocol_description : str, optional
        A description of the screen protocol; may contain very detailed
        information to reproduce some of that found in a screening database.
    protocol_identifier : str, optional
        A pointer to an externally defined protocol, usually in a screening
        database.
    reagent_set_description : str, optional
        A description of the set of reagents; may contain very detailed
        information to reproduce some of that information found in a screening
        database.
    reagent_set_identifier : str, optional
        A pointer to an externally defined set of reagents, usually in a
        screening database/automation database.
    reagents : Reagent, optional
    type : str, optional
        A human readable identifier for the screen type; e.g. RNAi, cDNA,
        SiRNA, etc. This string is likely to become an enumeration in future
        releases.
    """

    id: ScreenID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    description: Optional[str] = None
    name: Optional[str] = None
    plate_ref: List[PlateRef] = Field(default_factory=list)
    protocol_description: Optional[str] = None
    protocol_identifier: Optional[str] = None
    reagent_set_description: Optional[str] = None
    reagent_set_identifier: Optional[str] = None
    reagents: List[Reagent] = Field(default_factory=list)
    type: Optional[str] = None



from typing import Any, Dict, List, Optional

from pydantic import Field, root_validator

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .detector import Detector
from .dichroic import Dichroic
from .filter import Filter
from .filter_set import FilterSet
from .light_source_group import LightSourceGroupType
from .microscope import Microscope
from .objective import Objective
from .simple_types import InstrumentID


class Instrument(OMEType):
    """This element describes the instrument used to capture the Image.

    It is primarily a container for manufacturer's model and catalog numbers for
    the Microscope, LightSource, Detector, Objective and Filters components. The
    Objective element contains the additional elements LensNA and Magnification.
    The Filters element can be composed either of separate excitation, emission
    filters and a dichroic mirror or a single filter set. Within the Image itself,
    a reference is made to this one Filter element. There may be multiple light
    sources, detectors, objectives and filters on a microscope. Each of these has
    their own ID attribute, which can be referred to from Channel. It is
    understood that the light path configuration can be different for each
    channel, but cannot be different for each timepoint or each plane of an XYZ
    stack.

    Parameters
    ----------
    id : InstrumentID
    annotation_ref : AnnotationRef, optional
    detectors : Detector, optional
    dichroics : Dichroic, optional
    filter_sets : FilterSet, optional
    filters : Filter, optional
    light_source_group : List[LightSourceGroupType], optional
    microscope : Microscope, optional
    objectives : Objective, optional
    """

    id: InstrumentID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    detectors: List[Detector] = Field(default_factory=list)
    dichroics: List[Dichroic] = Field(default_factory=list)
    filter_sets: List[FilterSet] = Field(default_factory=list)
    filters: List[Filter] = Field(default_factory=list)
    light_source_group: List[LightSourceGroupType] = Field(default_factory=list)
    microscope: Optional[Microscope] = None
    objectives: List[Objective] = Field(default_factory=list)

    @root_validator(pre=True)
    def _root(cls, value: Dict[str, Any]):
        light_sources = {i.snake_name() for i in LightSourceGroupType.__args__}  # type: ignore
        lights = []
        for key in list(value):
            kind = {"kind": key}
            if key in light_sources:
                val = value.pop(key)
                if isinstance(val, dict):
                    lights.append({**val, **kind})
                elif isinstance(val, list):
                    lights.extend({**v, **kind} for v in val)
        if lights:
            value.setdefault("light_source_group", [])
            value["light_source_group"].extend(lights)
        return value



from typing import Optional

from ome_types._base_type import OMEType

from .settings import Settings
from .simple_types import (
    Binning,
    DetectorID,
    PositiveInt,
    UnitsElectricPotential,
    UnitsFrequency,
)


class DetectorSettings(Settings, OMEType):
    """This holds the setting applied to a detector as well as a reference to the
    detector.

    The ID is the detector used in this case. The values stored in
    DetectorSettings represent the variable values, fixed values not modified
    during the acquisition go in Detector.

    Each attribute now has an indication of what type of detector it applies to.
    This is preparatory work for cleaning up and possibly splitting this object
    into sub-types.

    Parameters
    ----------
    id : DetectorID
    binning : Binning, optional
        Represents the number of pixels that are combined to form larger
        pixels. {used:CCD,EMCCD}
    gain : float, optional
        The Gain of the detector. {used:CCD,EMCCD,PMT}
    integration : PositiveInt, optional
        This is the number of sequential frames that get averaged, to improve
        the signal-to-noise ratio. {used:CCD,EMCCD}
    offset : float, optional
        The Offset of the detector. {used:CCD,EMCCD}
    read_out_rate : float, optional
        The speed at which the detector can count pixels.  {used:CCD,EMCCD}
        This is the bytes per second that can be read from the detector (like
        a baud rate). Units are set by ReadOutRateUnit.
    read_out_rate_unit : UnitsFrequency, optional
        The units of the ReadOutRate - default:megahertz.
    voltage : float, optional
        The Voltage of the detector. {used:PMT} Units are set by VoltageUnit.
    voltage_unit : UnitsElectricPotential, optional
        The units of the Voltage of the detector - default:volts
    zoom : float, optional
        The Zoom or "Confocal Zoom" or "Scan Zoom" for a detector. {used:PMT}
    """

    id: DetectorID
    binning: Optional[Binning] = None
    gain: Optional[float] = None
    integration: Optional[PositiveInt] = None
    offset: Optional[float] = None
    read_out_rate: Optional[float] = None
    read_out_rate_unit: Optional[UnitsFrequency] = UnitsFrequency("MHz")
    voltage: Optional[float] = None
    voltage_unit: Optional[UnitsElectricPotential] = UnitsElectricPotential("V")
    zoom: Optional[float] = None



from enum import Enum
from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .manufacturer_spec import ManufacturerSpec
from .simple_types import ObjectiveID, UnitsLength


class Correction(Enum):
    """The correction applied to the lens."""

    ACHRO = "Achro"
    ACHROMAT = "Achromat"
    APO = "Apo"
    FL = "Fl"
    FLUAR = "Fluar"
    FLUOR = "Fluor"
    FLUOTAR = "Fluotar"
    NEOFLUAR = "Neofluar"
    OTHER = "Other"
    PLAN_APO = "PlanApo"
    PLAN_FLUOR = "PlanFluor"
    PLAN_NEOFLUAR = "PlanNeofluar"
    SUPER_FLUOR = "SuperFluor"
    UV = "UV"
    VIOLET_CORRECTED = "VioletCorrected"


class Immersion(Enum):
    """The immersion medium the lens is designed for."""

    AIR = "Air"
    GLYCEROL = "Glycerol"
    MULTI = "Multi"
    OIL = "Oil"
    OTHER = "Other"
    WATER = "Water"
    WATER_DIPPING = "WaterDipping"


class Objective(ManufacturerSpec, OMEType):
    """A description of the microscope's objective lens.

    Required elements include the lens numerical aperture, and the magnification,
    both of which a floating point (real) numbers. The values are those that are
    fixed for a particular objective: either because it has been manufactured to
    this specification or the value has been measured on this particular
    objective. Correction: This is the type of correction coating applied to this
    lens. Immersion: This is the types of immersion medium the lens is designed to
    work with. It is not the same as 'Medium' in ObjectiveRef (a single type) as
    here Immersion can have compound values like 'Multi'. LensNA: The numerical
    aperture of the lens (as a float) NominalMagnification: The specified
    magnification e.g. x10 CalibratedMagnification: The measured magnification
    e.g. x10.3 WorkingDistance: WorkingDistance of the lens.

    Parameters
    ----------
    id : ObjectiveID
    annotation_ref : AnnotationRef, optional
    calibrated_magnification : float, optional
        The magnification of the lens as measured by a calibration process-
        i.e. '59.987' for a 60X lens.
    correction : Correction, optional
        The correction applied to the lens
    immersion : Immersion, optional
        The immersion medium the lens is designed for
    iris : bool, optional
        Records whether or not the objective was fitted with an Iris.
    lens_na : float, optional
        The numerical aperture of the lens expressed as a floating point
        (real) number. Expected range 0.02 - 1.5
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    nominal_magnification : float, optional
        The magnification of the lens as specified by the manufacturer - i.e.
        '60' is a 60X lens. Note: The type of this has been changed from int
        to float to allow the specification of additional lenses e.g. 0.5X
        lens
    serial_number : str, optional
        The serial number of the component.
    working_distance : float, optional
        The working distance of the lens expressed as a floating point (real)
        number. Units are set by WorkingDistanceUnit.
    working_distance_unit : UnitsLength, optional
        The units of the working distance - default:microns.
    """

    id: ObjectiveID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    calibrated_magnification: Optional[float] = None
    correction: Optional[Correction] = None
    immersion: Optional[Immersion] = None
    iris: Optional[bool] = None
    lens_na: Optional[float] = None
    nominal_magnification: Optional[float] = None
    working_distance: Optional[float] = None
    working_distance_unit: Optional[UnitsLength] = UnitsLength("µm")



from typing import Any

from ._units import ureg

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

try:
    from . import model
    from .model import OME
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "Could not import 'ome_types.model.OME'.\nIf you are in a dev environment, "
        "you may need to run 'python -m src.ome_autogen'" + str(e)
    ) from None

from ._convenience import (  # isort:skip
    from_tiff,
    from_xml,
    to_dict,
    to_xml,
    validate_xml,
)

__all__ = [
    "from_tiff",
    "from_xml",
    "model",
    "OME",
    "to_dict",
    "to_xml",
    "ureg",
    "validate_xml",
    "__version__",
]


def __getattr__(name: str) -> Any:
    if name == "validate":
        import warnings

        warnings.warn(
            "'ome_types.validate' has been renamed to 'ome_types.validate_xml. "
            "This will raise an exception in the future.",
            FutureWarning,
            stacklevel=2,
        )
        return validate_xml
    raise AttributeError("module {__name__!r} has no attribute {name!r}")



from __future__ import annotations

import builtins
import os
import re
import shutil
from dataclasses import dataclass, field
from itertools import chain
from pathlib import Path
from textwrap import dedent, indent, wrap
from typing import Any, DefaultDict, Generator, Iterable, Iterator

import black
import isort.api
from autoflake import fix_code
from numpydoc.docscrape import NumpyDocString, Parameter
from xmlschema import XMLSchema
from xmlschema.validators import (
    XsdAnyAttribute,
    XsdAnyElement,
    XsdAtomicBuiltin,
    XsdAttribute,
    XsdComponent,
    XsdElement,
    XsdType,
)

try:
    # xmlschema ≥ v1.4.0
    from xmlschema import names as qnames
except ImportError:
    from xmlschema import qnames


# Track all camel-to-snake and pluralization results so we can include them in the model
camel_snake_registry: dict[str, str] = {}
plural_registry: dict[tuple[str, str], str] = {}
LISTS: DefaultDict[str, set[str]] = DefaultDict(set)

# FIXME: Work out a better way to implement these override hacks.


@dataclass
class Override:
    type_: str
    default: str | None = None
    imports: str | None = None
    body: str | None = None

    def __post_init__(self) -> None:
        if self.imports:
            self.imports = dedent(self.imports)
        if self.body:
            self.body = indent(dedent(self.body), " " * 4)


@dataclass
class ClassOverride:
    base_type: str | None = None
    imports: str | None = None
    fields: str | None = None
    fields_suppress: set[str] = field(default_factory=set)
    body: str | None = None

    def __post_init__(self) -> None:
        if self.imports:
            self.imports = dedent(self.imports)
        if self.fields:
            self.fields = indent(dedent(self.fields), " " * 4)
        if self.body:
            self.body = indent(dedent(self.body), " " * 4)


# Maps XSD TypeName to Override configuration, used to control output for that type.
OVERRIDES = {
    "MetadataOnly": Override(type_="bool", default="False"),
    # FIXME: Type should be xml.etree.ElementTree.Element but isinstance checks
    # with that class often mysteriously fail so the validator fails.
    "XMLAnnotation/Value": Override(type_="Element"),
    "BinData/Length": Override(type_="int"),
    # FIXME: hard-coded subclass lists
    "Instrument/LightSourceGroup": Override(
        type_="List[LightSourceGroupType]",
        default="Field(default_factory=list)",
        imports="""
            from typing import Dict, Union, Any
            from pydantic import root_validator
            from .light_source_group import LightSourceGroupType
        """,
        body="""
            @root_validator(pre=True)
            def _root(cls, value: Dict[str, Any]):
                light_sources = {i.snake_name() for i in LightSourceGroupType.__args__}  # type: ignore
                lights = []
                for key in list(value):
                    kind = {"kind": key}
                    if key in light_sources:
                        val = value.pop(key)
                        if isinstance(val, dict):
                            lights.append({**val, **kind})
                        elif isinstance(val, list):
                            lights.extend({**v, **kind} for v in val)
                if lights:
                    value.setdefault("light_source_group", [])
                    value["light_source_group"].extend(lights)
                return value
        """,
    ),
    "ROI/Union": Override(
        type_="List[ShapeGroupType]",
        default="Field(default_factory=list)",
        imports="""
            from typing import Dict, Union, Any, Sequence, Iterator
            from pydantic import validator

            from .annotation_ref import AnnotationRef
            from .shape_group import ShapeGroupType
            from .simple_types import ROIID
        """,
        body="""
            @validator("union", pre=True)
            def _validate_union(cls, value: Any) -> Sequence[Dict[str, Any]]:
                if isinstance(value, dict):
                    return list(cls._flatten_union_dict(value))
                if not isinstance(value, Sequence):
                    raise TypeError("must be dict or sequence of dicts")
                return value

            @classmethod
            def _flatten_union_dict(cls, nested: Dict[str, Any], keyname: str = "kind"
            ) -> Iterator[Dict[str, Any]]:
                for key, value in nested.items():
                    keydict = {keyname: key} if keyname else {}
                    if isinstance(value, list):
                        yield from ({**x, **keydict} for x in value)
                    else:
                        yield {**value, **keydict}
        """,
    ),
    "OME/StructuredAnnotations": Override(
        type_="List[Annotation]",
        default="Field(default_factory=list)",
        imports="""
            from typing import Dict, Union, Any
            from pydantic import validator
            from .annotation import Annotation
            from .boolean_annotation import BooleanAnnotation
            from .comment_annotation import CommentAnnotation
            from .double_annotation import DoubleAnnotation
            from .file_annotation import FileAnnotation
            from .list_annotation import ListAnnotation
            from .long_annotation import LongAnnotation
            from .map_annotation import MapAnnotation
            from .tag_annotation import TagAnnotation
            from .term_annotation import TermAnnotation
            from .timestamp_annotation import TimestampAnnotation
            from .xml_annotation import XMLAnnotation

            _annotation_types: Dict[str, type] = {
                "boolean_annotation": BooleanAnnotation,
                "comment_annotation": CommentAnnotation,
                "double_annotation": DoubleAnnotation,
                "file_annotation": FileAnnotation,
                "list_annotation": ListAnnotation,
                "long_annotation": LongAnnotation,
                "map_annotation": MapAnnotation,
                "tag_annotation": TagAnnotation,
                "term_annotation": TermAnnotation,
                "timestamp_annotation": TimestampAnnotation,
                "xml_annotation": XMLAnnotation,
            }
        """,
        body="""
            @validator("structured_annotations", pre=True, each_item=True)
            def validate_structured_annotations(
                cls, value: Union[Annotation, Dict[Any, Any]]
            ) -> Annotation:
                if isinstance(value, Annotation):
                    return value
                elif isinstance(value, dict):
                    try:
                        _type = value.pop("_type")
                    except KeyError:
                        raise ValueError(
                            "dict initialization requires _type"
                        ) from None
                    try:
                        annotation_cls = _annotation_types[_type]
                    except KeyError:
                        raise ValueError(f"unknown Annotation type '{_type}'") from None
                    return annotation_cls(**value)
                else:
                    raise ValueError("invalid type for annotation values")
        """,
    ),
    "TiffData/UUID": Override(
        type_="Optional[UUID]",
        default="None",
        imports="""
            from typing import Optional
            from .simple_types import UniversallyUniqueIdentifier
            from ome_types._base_type import OMEType

            class UUID(OMEType):
                file_name: str
                value: UniversallyUniqueIdentifier
        """,
    ),
    "M/K": Override(type_="str", default=""),
}


# Maps XSD TypeName to ClassOverride configuration, used to control dataclass
# generation.
CLASS_OVERRIDES = {
    "OME": ClassOverride(
        imports="""
            from typing import Any
            import weakref
            from ome_types import util
            from pathlib import Path
        """,
        body="""
            def __init__(self, **data: Any) -> None:
                super().__init__(**data)
                self._link_refs()

            def _link_refs(self) -> None:
                ids = util.collect_ids(self)
                for ref in util.collect_references(self):
                    ref._ref = weakref.ref(ids[ref.id])

            def __setstate__(self: Any, state: Dict[str, Any]) -> None:
                '''Support unpickle of our weakref references.'''
                super().__setstate__(state)
                self._link_refs()

            @classmethod
            def from_xml(cls, xml: Union[Path, str]) -> 'OME':
                from ome_types import from_xml
                return from_xml(xml)

            @classmethod
            def from_tiff(cls, path: Union[Path, str]) -> 'OME':
                from ome_types import from_tiff
                return from_tiff(path)

            def to_xml(self) -> str:
                from ome_types import to_xml
                return to_xml(self)
        """,
    ),
    "Reference": ClassOverride(
        imports="""
            from pydantic import Field
            from typing import Any, Optional, TYPE_CHECKING
            from weakref import ReferenceType
            from .simple_types import LSID
        """,
        fields="""
            if TYPE_CHECKING:
                _ref: Optional["ReferenceType[OMEType]"]

            id: LSID
            _ref = None
        """,
        # FIXME Could make `ref` abstract and implement stronger-typed overrides
        # in subclasses.
        body="""
            @property
            def ref(self) -> Any:
                if self._ref is None:
                    raise ValueError("references not yet resolved on root OME object")
                return self._ref()
        """,
    ),
    "XMLAnnotation": ClassOverride(
        imports="""
            from xml.etree import ElementTree
            from typing import Generator
            from typing import Callable, Generator, Any, Dict

            class Element(ElementTree.Element):
                '''ElementTree.Element that supports pydantic validation.'''
                @classmethod
                def __get_validators__(cls) -> Generator[Callable[[Any], Any], None, None]:
                    yield cls.validate

                @classmethod
                def validate(cls, v: Any) -> ElementTree.Element:
                    if isinstance(v, ElementTree.Element):
                        return v
                    try:
                        return ElementTree.fromstring(v)
                    except ElementTree.ParseError as e:
                        raise ValueError(f"Invalid XML string: {e}")
        """,
        body="""
            # NOTE: pickling this object requires xmlschema>=1.4.1

            def dict(self, **k: Any) -> Dict[str, Any]:
                d = super().dict(**k)
                d["value"] = ElementTree.tostring(
                    d.pop("value"), encoding="unicode", method="xml"
                ).strip()
                return d
        """,
    ),
    "BinData": ClassOverride(base_type="object", fields="value: str"),
    "Map": ClassOverride(fields_suppress={"K"}),
    "M": ClassOverride(base_type="object", fields="value: str"),
    "LightEmittingDiode": ClassOverride(
        fields='kind: Literal["light_emitting_diode"] = "light_emitting_diode"',
        imports="from typing_extensions import Literal",
    ),
    "Laser": ClassOverride(
        fields='kind: Literal["laser"] = "laser"',
        imports="from typing_extensions import Literal",
    ),
    "Arc": ClassOverride(
        fields='kind: Literal["arc"] = "arc"',
        imports="from typing_extensions import Literal",
    ),
    "Filament": ClassOverride(
        fields='kind: Literal["filament"] = "filament"',
        imports="from typing_extensions import Literal",
    ),
    "GenericExcitationSource": ClassOverride(
        fields='kind: Literal["generic_excitation_source"] = "generic_excitation_source"',
        imports="from typing_extensions import Literal",
    ),
    "Label": ClassOverride(
        fields='kind: Literal["label"] = "label"',
        imports="from typing_extensions import Literal",
    ),
    "Point": ClassOverride(
        fields='kind: Literal["point"] = "point"',
        imports="from typing_extensions import Literal",
    ),
    "Mask": ClassOverride(
        fields='kind: Literal["mask"] = "mask"',
        imports="from typing_extensions import Literal",
    ),
    "Rectangle": ClassOverride(
        fields='kind: Literal["rectangle"] = "rectangle"',
        imports="from typing_extensions import Literal",
    ),
    "Polygon": ClassOverride(
        fields='kind: Literal["polygon"] = "polygon"',
        imports="from typing_extensions import Literal",
    ),
    "Polyline": ClassOverride(
        fields='kind: Literal["polyline"] = "polyline"',
        imports="from typing_extensions import Literal",
    ),
    "Line": ClassOverride(
        fields='kind: Literal["line"] = "line"',
        imports="from typing_extensions import Literal",
    ),
    "Ellipse": ClassOverride(
        fields='kind: Literal["ellipse"] = "ellipse"',
        imports="from typing_extensions import Literal",
    ),
}


def autoflake(text: str, **kwargs: Any) -> str:
    kwargs.setdefault("remove_all_unused_imports", True)
    kwargs.setdefault("remove_unused_variables", True)
    return fix_code(text, **kwargs)


def black_format(text: str, line_length: int = 88) -> str:
    return black.format_str(text, mode=black.FileMode(line_length=line_length))


def sort_imports(text: str) -> str:
    return isort.api.sort_code_string(text, profile="black", float_to_top=True)


def sort_types(el: XsdType) -> str:
    if not el.is_complex() and not el.base_type.is_restriction():
        return "    " + el.local_name.lower()
    return el.local_name.lower()


def sort_prop(prop: Member) -> str:
    return ("" if prop.default_val_str else "   ") + prop.format().lower()


def as_identifier(s: str) -> str:
    # Remove invalid characters
    _s = re.sub("[^0-9a-zA-Z_]", "", s)
    # Remove leading characters until we find a letter or underscore
    _s = re.sub("^[^a-zA-Z_]+", "", _s)
    if not _s:
        raise ValueError(f"Could not clean {s}: nothing left")
    return _s


CAMEL_SNAKE_OVERRIDES = {"ROIs": "rois"}


def camel_to_snake(name: str) -> str:
    result = CAMEL_SNAKE_OVERRIDES.get(name, None)
    if not result:
        # https://stackoverflow.com/a/1176023
        result = re.sub("([A-Z]+)([A-Z][a-z]+)", r"\1_\2", name)
        result = re.sub("([a-z0-9])([A-Z])", r"\1_\2", result)
        result = result.lower().replace(" ", "_")
    camel_snake_registry[name] = result
    return result


def local_import(item_type: str) -> str:
    return f"from .{camel_to_snake(item_type)} import {item_type}"


def get_docstring(component: XsdComponent | XsdType, summary: bool = False) -> str:
    try:
        doc = dedent(component.annotation.documentation[0].text).strip()
        # some docstrings include a start ('*Word') which makes sphinx angry
        # this line wraps those in backticks
        doc = re.sub(r"(\*\w+)\s", r"`\1` ", doc)
        # make sure the first line is followed by a double newline
        # and preserve paragraphs
        if summary:
            doc = re.sub(r"\.\s", ".\n\n", doc, count=1)
        # textwrap each paragraph seperately
        paragraphs = ["\n".join(wrap(p.strip(), width=78)) for p in doc.split("\n\n")]
        # join and return
        return "\n\n".join(paragraphs)
    except (AttributeError, IndexError):
        return ""


def make_dataclass(component: XsdComponent | XsdType) -> list[str]:
    class_override = CLASS_OVERRIDES.get(component.local_name, None)
    lines = ["from ome_types._base_type import OMEType", ""]
    if isinstance(component, XsdType):
        base_type = component.base_type
    else:
        base_type = component.type.base_type

    if class_override and class_override.base_type:
        if class_override.base_type == "object":
            base_name = "(OMEType)"
        else:
            base_name = f"({class_override.base_type}, OMEType)"
        base_type = None
    elif base_type and not hasattr(base_type, "python_type"):
        base_name = f"({base_type.local_name}, OMEType)"
        if base_type.is_complex():
            lines += [local_import(base_type.local_name)]
        else:
            lines += [f"from .simple_types import {base_type.local_name}"]
    else:
        base_name = "(OMEType)"
    if class_override and class_override.imports:
        lines.append(class_override.imports)

    base_members = set()
    _basebase = base_type
    while _basebase:
        base_members.update(set(iter_members(base_type)))
        _basebase = _basebase.base_type
    skip_names = set()
    if class_override:
        skip_names.update(class_override.fields_suppress)

    members = MemberSet(
        m
        for m in iter_members(component)
        if m not in base_members and m.local_name not in skip_names
    )
    lines += members.imports()
    lines += members.locals()

    lines += [f"class {component.local_name}{base_name}:"]
    doc = get_docstring(component, summary=True)
    doc = MemberSet(iter_members(component)).docstring(
        doc or f"{component.local_name}."
    )
    doc = f'"""{doc.strip()}\n"""\n'
    lines += indent(doc, "    ").splitlines()
    if class_override and class_override.fields:
        lines.append(class_override.fields)
    lines += members.lines(indent=1)

    if class_override and class_override.body:
        lines.append(class_override.body)
    lines += members.body()

    return lines


def make_abstract_class(component: XsdComponent) -> list[str]:
    # FIXME: ? this might be a bit of an OME-schema-specific hack
    # this seems to be how abstract is used in the OME schema
    for e in component.iter_components():
        if e != component:
            raise NotImplementedError(
                "Don't yet know how to handle abstract class with sub-components"
            )

    subs = [
        el
        for el in component.schema.elements.values()
        if el.substitution_group == component.name
    ]

    if not subs:
        raise NotImplementedError(
            "Don't know how to handle abstract class without substitutionGroups"
        )

    for el in subs:
        if not el.type.is_extension() and el.type.base_type == component.type:
            raise NotImplementedError(
                "Expected all items in substitution group to extend "
                f"the type {component.type} of Abstract element {component}"
            )

    sub_names = [el.local_name for el in subs]
    lines = ["from typing import Union", *[local_import(n) for n in sub_names]]
    lines += [local_import(component.type.local_name)]
    lines += [f"{component.local_name} = {component.type.local_name}", ""]
    lines += [f"{component.local_name}Type = Union[{', '.join(sub_names)}]"]
    return lines


def make_enum(component: XsdComponent) -> list[str]:
    name = component.local_name
    _type = component.type if hasattr(component, "type") else component
    if _type.is_list():
        _type = _type.item_type
    lines = ["from enum import Enum", ""]
    lines += [f"class {name}(Enum):"]
    doc = get_docstring(component, summary=True)
    if doc:
        if not doc.endswith("."):
            doc += "."
        doc = f'"""{doc}\n"""\n'
        lines += indent(doc, "    ").splitlines()
    enum_elems = list(_type.elem.iter("enum"))
    facets = _type.get_facet(qnames.XSD_ENUMERATION)
    members: list[tuple[str, str]] = []
    if enum_elems:
        for el, value in zip(enum_elems, facets.enumeration):
            _name = el.attrib["enum"]
            if _type.base_type.python_type.__name__ == "str":
                value = f'"{value}"'
            members.append((_name, value))
    else:
        for e in facets.enumeration:
            members.append((camel_to_snake(e), repr(e)))

    for n, v in sorted(members):
        lines.append(f"    {as_identifier(n).upper()} = {v}")
    return lines


def make_color() -> list[str]:
    color = """
    from pydantic import color
    class Color(color.Color):
        def __init__(self, val: color.ColorType) -> None:
            if isinstance(val, int):
                val = self._int2tuple(val)
            super().__init__(val)

        @classmethod
        def _int2tuple(cls, val: int):
            return (val >> 24 & 255, val >> 16 & 255, val >> 8 & 255, (val & 255) / 255)

        def as_int32(self) -> int:
            r, g, b, *a = self.as_rgb_tuple()
            v = r << 24 | g << 16 | b << 8 | int((a[0] if a else 1) * 255)
            if v < 2 ** 32 // 2:
                return v
            return v - 2 ** 32

        def __eq__(self, o: object) -> bool:
            if isinstance(o, Color):
                return self.as_int32() == o.as_int32()
            return False

        def __int__(self) -> int:
            return self.as_int32()
    """
    return dedent(color).strip().splitlines()


facet_converters = {
    qnames.XSD_PATTERN: lambda f: [f"regex = re.compile(r'{f.regexps[0]}')"],
    qnames.XSD_MIN_INCLUSIVE: lambda f: [f"ge = {f.value}"],
    qnames.XSD_MIN_EXCLUSIVE: lambda f: [f"gt = {f.value}"],
    qnames.XSD_MAX_INCLUSIVE: lambda f: [f"le = {f.value}"],
    qnames.XSD_MAX_EXCLUSIVE: lambda f: [f"lt = {f.value}"],
    qnames.XSD_LENGTH: lambda f: [f"min_length = {f.value}", f"max_length = {f.value}"],
    qnames.XSD_MIN_LENGTH: lambda f: [f"min_length = {f.value}"],
    qnames.XSD_MAX_LENGTH: lambda f: [f"max_length = {f.value}"],
}


def iter_all_members(
    component: XsdComponent,
) -> Generator[XsdElement | XsdAttribute, None, None]:
    for c in component.iter_components((XsdElement, XsdAttribute)):
        if c is component:
            continue
        yield c


def iter_members(
    component: XsdElement | XsdType,
) -> Generator[XsdElement | XsdAttribute, None, None]:
    if isinstance(component, XsdElement):
        for attr in component.attributes.values():
            if isinstance(attr, XsdAttribute):
                yield attr
        yield from component.iterchildren()
    else:
        yield from iter_all_members(component)


def is_enum_type(obj: XsdType) -> bool:
    """Return true if XsdType represents an enumeration."""
    return obj.get_facet(qnames.XSD_ENUMERATION) is not None


class Member:
    def __init__(self, component: XsdElement | XsdAttribute):
        self.component = component
        if component.is_global():
            raise ValueError(f"global component {component!r} not allowed")

    @property
    def identifier(self) -> str:
        if isinstance(self.component, (XsdAnyElement, XsdAnyAttribute)):
            return self.component.local_name
        name = camel_to_snake(self.component.local_name)
        if self.plural:
            plural = camel_to_snake(self.plural)
            plural_registry[(self.parent_name, name)] = plural
            name = plural
        if not name.isidentifier():
            raise ValueError(f"failed to make identifier of {self!r}")
        return name

    @property
    def plural(self) -> str | None:
        """Plural form of component name, if available."""
        if (
            isinstance(self.component, XsdElement)
            and self.component.is_multiple()
            and self.component.ref
            and self.component.ref.annotation
        ):
            appinfo = self.component.ref.annotation.appinfo
            if len(appinfo) != 1:
                raise ValueError("unexpected multiple appinfo elements")
            plural = appinfo[0].find("xsdfu/plural")
            if plural is not None:
                return plural.text
        return None

    @property
    def type(self) -> XsdType:
        return self.component.type

    def to_numpydoc_param(self) -> Parameter:
        _type = self.type_string
        _type += ", optional" if self.is_optional else ""
        desc = get_docstring(self.component)
        desc = re.sub(r"\s?\[.+\]", "", desc)  # remove bracketed types
        return Parameter(self.identifier, _type, wrap(desc))

    @property
    def is_builtin_type(self) -> bool:
        return isinstance(self.type, XsdAtomicBuiltin)

    @property
    def is_decimal(self) -> bool:
        return self.component.type.is_derived(
            self.component.schema.builtin_types()["decimal"]
        )

    @property
    def is_nonref_id(self) -> bool:
        """Return True for 'id' fields that aren't part of a Reference type."""
        if self.identifier != "id":
            return False
        # Walk up the containment tree until we find something with a base_type.
        p = self.component.parent
        while p is not None and not hasattr(p, "base_type"):
            p = p.parent
        if p is not None:
            # Walk the type hierarchy looking for 'Reference'.
            pt = p.base_type
            while pt is not None:
                if pt.local_name == "Reference":
                    return False
                pt = pt.base_type
        # If we get here, we have an 'id' that isn't in a Reference type.
        return True

    @property
    def is_ref_id(self) -> bool:
        if self.identifier == "id":
            return not self.is_nonref_id
        return False

    @property
    def parent_name(self) -> str:
        """Local name of component's first named ancestor."""
        p = self.component.parent
        while not p.local_name and p.parent is not None:
            p = p.parent
        return p.local_name

    @property
    def key(self) -> str:
        name = f"{self.parent_name}/{self.component.local_name}"
        if name not in OVERRIDES and self.component.local_name in OVERRIDES:
            return self.component.local_name
        return name

    def locals(self) -> list[str]:
        if self.key in OVERRIDES:
            return []
        if isinstance(self.component, (XsdAnyElement, XsdAnyAttribute)):
            return []
        if not self.type or self.type.is_global():
            return []
        locals_: list[str] = []
        # FIXME: this bit is mostly hacks
        if self.type.is_complex() and self.component.ref is None:
            locals_.append("\n".join(make_dataclass(self.component)) + "\n")
        if self.type.is_restriction() and is_enum_type(self.type):
            locals_.append("\n".join(make_enum(self.component)) + "\n")
        if self.type.is_list() and is_enum_type(self.type.item_type):
            locals_.append("\n".join(make_enum(self.component)) + "\n")
        return locals_

    def imports(self) -> list[str]:
        if self.key in OVERRIDES:
            _imp = OVERRIDES[self.key].imports
            return [_imp] if _imp else []
        if isinstance(self.component, (XsdAnyElement, XsdAnyAttribute)):
            return ["from typing import Any"]
        imports = []
        if not self.max_occurs:
            imports.append("from typing import List")
            if self.is_optional:
                imports.append("from pydantic import Field")
        elif self.is_optional:
            imports.append("from typing import Optional")
        if self.is_decimal:
            imports.append("from typing import cast")
        if self.type.is_datetime():
            imports.append("from datetime import datetime")
        if not self.is_builtin_type and self.type.is_global():
            # FIXME: hack
            if not self.type.local_name == "anyType":
                if self.type.is_complex():
                    imports.append(local_import(self.type.local_name))
                else:
                    imports.append(f"from .simple_types import {self.type.local_name}")

        if self.component.ref is not None:
            if self.component.ref.local_name not in OVERRIDES:
                imports.append(local_import(self.component.ref.local_name))

        return imports

    def body(self) -> str:
        if self.key in OVERRIDES:
            return OVERRIDES[self.key].body or ""
        return ""

    @property
    def type_string(self) -> str:
        """Single type, without Optional, etc..."""
        if self.key in OVERRIDES:
            return OVERRIDES[self.key].type_
        if isinstance(self.component, (XsdAnyElement, XsdAnyAttribute)):
            return "Any"
        if self.component.ref is not None:
            if not self.component.ref.is_global():
                raise ValueError("local ref not supported")
            return self.component.ref.local_name

        if self.type.is_datetime():
            return "datetime"
        if self.is_builtin_type:
            return self.type.python_type.__name__

        if self.type.is_global():
            return self.type.local_name
        elif self.type.is_complex() or self.type.is_list():
            return self.component.local_name

        if self.type.is_restriction():
            # enumeration
            enum = self.type.get_facet(qnames.XSD_ENUMERATION)
            if enum:
                return self.component.local_name
            if self.type.base_type.local_name == "string":
                return "str"
        return ""

    @property
    def full_type_string(self) -> str:
        """Full type, like Optional[List[str]]."""
        if self.key in OVERRIDES and self.type_string:
            return f": {self.type_string}"
        type_string = self.type_string
        if not type_string:
            return ""
        if not self.max_occurs:
            LISTS[self.parent_name].add(self.component.local_name)

            type_string = f"List[{type_string}]"
        elif self.is_optional:
            type_string = f"Optional[{type_string}]"
        return f": {type_string}" if type_string else ""

    @property
    def default_val_str(self) -> str:
        if self.key in OVERRIDES:
            default = OVERRIDES[self.key].default
            return f" = {default}" if default else ""
        elif not self.is_optional:
            return ""

        if not self.max_occurs:
            default_val = "Field(default_factory=list)"
        else:
            default_val = self.component.default
            if default_val is not None:
                if is_enum_type(self.type):
                    default_val = f"{self.type_string}('{default_val}')"
                elif hasattr(builtins, self.type_string):
                    default_val = repr(getattr(builtins, self.type_string)(default_val))
                if self.type_string == "Color":
                    default_val = "Color('white')"
                elif self.is_decimal:
                    default_val = f"cast({self.type_string}, {default_val})"
            else:
                default_val = "None"
        return f" = {default_val}"

    @property
    def max_occurs(self) -> int | None:
        default = None if self.type.is_list() else 1
        return getattr(self.component, "max_occurs", default)

    @property
    def is_optional(self) -> bool:
        if self.is_ref_id:
            return False
        if getattr(self.component.parent, "model", "") == "choice":
            return True
        if hasattr(self.component, "min_occurs"):
            return self.component.min_occurs == 0
        return self.component.is_optional()

    def __repr__(self) -> str:
        type_ = "element" if isinstance(self.component, XsdElement) else "attribute"
        return f"<Member {type_} {self.component.local_name}>"

    def format(self) -> str:
        return f"{self.identifier}{self.full_type_string}{self.default_val_str}"


class MemberSet:
    def __init__(self, initial: Iterable[Member] = ()):
        # Use a list to maintain insertion order.
        self._members: list[Member] = []
        self.update(initial)

    def add(self, member: Member) -> None:
        if not isinstance(member, Member):
            member = Member(member)
        # We don't expect very many elements so this O(n) check is fine.
        if member in self._members:
            return
        self._members.append(member)

    def update(self, members: Iterable[Member]) -> None:
        for member in members:
            self.add(member)

    def lines(self, indent: int = 1) -> list[str]:
        if not self._members:
            lines = ["    " * indent + "pass"]
        else:
            lines = [
                "    " * indent + m.format()
                for m in sorted(self._members, key=sort_prop)
            ]
        return lines

    def imports(self) -> list[str]:
        return list(chain.from_iterable(m.imports() for m in self._members))

    def locals(self) -> list[str]:
        return list(chain.from_iterable(m.locals() for m in self._members))

    def body(self) -> list[str]:
        return [m.body() for m in self._members]

    def has_non_default_args(self) -> bool:
        return any(not m.default_val_str for m in self._members)

    def has_nonref_id(self) -> bool:
        return any(m.is_nonref_id for m in self._members)

    @property
    def non_defaults(self) -> MemberSet:
        return MemberSet(m for m in self._members if not m.default_val_str)

    def __iter__(self) -> Iterator[Member]:
        return iter(self._members)

    def docstring(self, summary: str = "") -> str:
        ds = NumpyDocString(summary)
        ds["Parameters"] = [
            m.to_numpydoc_param() for m in sorted(self._members, key=sort_prop)
        ]
        return str(ds)


class GlobalElem:
    def __init__(self, elem: XsdElement | XsdType):
        if not elem.is_global():
            raise ValueError("Element must be global")
        self.elem = elem

    @property
    def type(self) -> XsdType:
        return self.elem if self.is_type else self.elem.type

    @property
    def is_complex(self) -> bool:
        if hasattr(self.type, "is_complex"):
            return self.type.is_complex()
        return False

    @property
    def is_element(self) -> bool:
        return isinstance(self.elem, XsdElement)

    @property
    def is_type(self) -> bool:
        return isinstance(self.elem, XsdType)

    @property
    def is_enum(self) -> bool:
        is_enum = bool(self.elem.get_facet(qnames.XSD_ENUMERATION) is not None)
        if is_enum:
            if not len(self.elem.facets) == 1:
                raise NotImplementedError("Unexpected enum with multiple facets")
        return is_enum

    def _simple_class(self) -> list[str]:
        if self.type.local_name == "Color":
            return make_color()
        if self.is_enum:
            return make_enum(self.elem)
        # Hack for xmlschema > 1.4.1
        if self.type.local_name == "base64Binary":
            return ["class base64Binary(ConstrainedStr):", "    pass"]
        if self.type.local_name == "Hex40":
            return [
                "class Hex40(ConstrainedStr):",
                "    min_length = 40",
                "    max_length = 40",
            ]

        lines = []
        if self.type.base_type.is_restriction():
            parent = self.type.base_type.local_name
        else:
            # it's a restriction of a builtin
            pytype = self.elem.base_type.python_type.__name__
            parent = f"Constrained{pytype.title()}"
            lines.extend([f"from pydantic.types import {parent}", ""])
        lines.append(f"class {self.elem.local_name}({parent}):")

        members = []
        for key, facet in self.elem.facets.items():
            members.extend([f"    {line}" for line in facet_converters[key](facet)])
        lines.extend(members if members else ["    pass"])
        if any("re.compile" in m for m in members):
            lines = ["import re", "", *lines]
        return lines

    def lines(self) -> str:
        if not self.is_complex:
            lines = self._simple_class()
        elif self.elem.abstract:
            lines = make_abstract_class(self.elem)
        else:
            lines = make_dataclass(self.elem)
        return "\n".join(lines)

    def format(self) -> str:
        return black_format(sort_imports(autoflake(self.lines() + "\n")))

    def write(self, filename: str) -> None:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, "w", encoding="utf-8") as f:
            f.write(self.format())

    @property
    def fname(self) -> str:
        return f"{camel_to_snake(self.elem.local_name)}.py"


_this_dir = os.path.dirname(__file__)
_url = os.path.join(_this_dir, "ome_types", "ome-2016-06.xsd")
_target = os.path.join(_this_dir, "ome_types", "model")

AnnotationID_Override = r"""class AnnotationID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Annotation:\S+)|(Annotation:\S+)"
    )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v) -> "AnnotationID":
        if not cls.regex.match(v):
            search = cls.regex.search(v)
            if search:
                import warnings

                new_v = search.group()
                warnings.warn(f"Casting invalid AnnotationID {v!r} to {new_v!r}")
                v = new_v
        return v
"""

_SIMPLE_OVERRIDES = {"AnnotationID": AnnotationID_Override}


def convert_schema(url: str = _url, target_dir: str = _target) -> None:
    print("Inspecting XML schema ...")
    if isinstance(url, Path):
        url = str(url)
    schema = XMLSchema(url)
    print("Building model ...")
    shutil.rmtree(target_dir, ignore_errors=True)
    init_imports = []
    simples: list[GlobalElem] = []
    for elem in sorted(schema.types.values(), key=sort_types):
        if elem.local_name in OVERRIDES:
            continue
        converter = GlobalElem(elem)
        if not elem.is_complex():
            simples.append(converter)
            continue
        targetfile = os.path.join(target_dir, converter.fname)
        init_imports.append((converter.fname, elem.local_name))
        converter.write(filename=targetfile)

    for elem in schema.elements.values():
        if elem.local_name in OVERRIDES:
            continue
        converter = GlobalElem(elem)
        targetfile = os.path.join(target_dir, converter.fname)
        init_imports.append((converter.fname, elem.local_name))
        converter.write(filename=targetfile)

    text = "\n".join(
        [_SIMPLE_OVERRIDES.get(s.elem.local_name) or s.format() for s in simples]
    )
    text = black_format(sort_imports(text))
    with open(os.path.join(target_dir, "simple_types.py"), "w", encoding="utf-8") as f:
        f.write(text)

    text = ""
    for _fname, classname in init_imports:
        text += local_import(classname) + "\n"
    text = sort_imports(text)
    text += f"\n\n__all__ = [{', '.join(sorted(repr(i[1]) for i in init_imports))}]"
    # FIXME These could probably live somewhere else less visible to end-users.
    if len(plural_registry) != len(set(plural_registry.values())):
        raise Exception(
            "singular-to-plural mapping is not invertible (duplicate plurals)"
        )
    text += "\n\n_singular_to_plural = " + repr(
        {k: plural_registry[k] for k in sorted(plural_registry)}
    )
    text += "\n\n_plural_to_singular = " + repr(
        {plural_registry[k]: k[1] for k in sorted(plural_registry)}
    )
    if len(camel_snake_registry) != len(set(camel_snake_registry.values())):
        raise Exception("camel-to-snake mapping is not invertible (duplicate snakes)")
    text += "\n\n_camel_to_snake = " + repr(
        {k: camel_snake_registry[k] for k in sorted(camel_snake_registry)}
    )
    text += "\n\n_snake_to_camel = " + repr(
        {camel_snake_registry[k]: k for k in sorted(camel_snake_registry)}
    )
    text += "\n\n_lists = " + repr(dict(LISTS))
    text = black_format(text)
    with open(os.path.join(target_dir, "__init__.py"), "w", encoding="utf-8") as f:
        f.write(text)


if __name__ == "__main__":
    # for testing
    convert_schema()



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import DatasetID


class DatasetRef(Reference, OMEType):
    """The DatasetRef element refers to a Dataset by specifying the Dataset ID
    attribute.

    One or more DatasetRef elements may be listed within the Image element to
    specify what Datasets the Image belongs to.

    Parameters
    ----------
    id : DatasetID
    """

    id: DatasetID
